from typing import Union, List

from vortexasdk.api.serdes import FromDictMixin
from vortexasdk.logger import get_logger

import pandas as pd

logger = get_logger(__name__)


def create_list(list_of_dicts, output_class: FromDictMixin) -> List:
    """Convert each list element into an instance of the output class."""
    logger.debug(f"Converting list of dictionaries to list of {output_class}")
    return [output_class.from_dict(d) for d in list_of_dicts]


def format_datatypes(df: pd.DataFrame) -> pd.DataFrame:
    """Format the relevant columns with sensible datatypes"""
    timestamp_cols = [col for col in df.columns if "timestamp" in col]

    for col in timestamp_cols:
        df[col] = pd.to_datetime(df[col])

    return df


def create_dataframe(
    columns: Union[None, List[str]],
    default_columns: List[str],
    data: List[dict],
    logger_description: str,
) -> pd.DataFrame:
    """
    :param columns: Columns to be used in the dataframe
    :param default_columns: Default columns to be used if columns is None
    :param data: records that will be present in the dataframe
    :param logger_description: name of the type of record created. Used for logging.
    :return: pd.DataFrame of records with specified columns
    """
    logger.debug(f"Creating DataFrame of {logger_description}")

    if columns is None:
        df = pd.DataFrame(data=data, columns=default_columns)
    elif columns == "all":
        df = pd.DataFrame(data=data)
    else:
        df = pd.DataFrame(data=data, columns=columns)

    return format_datatypes(df)



from unittest import TestCase

import jsons

from vortexasdk.api.product import ProductEntity


class TestProductEntity(TestCase):
    def test_serialize(self):
        with open("tests/api/examples/product_entity1.json", "r") as f:
            serialized = f.read()
            deserialized = jsons.loads(serialized, ProductEntity)

            expected = ProductEntity(
                id="6f11b0724c9a4e85ffa7f1445bc768f054af755a090118dcf99f14745c261653",
                layer="group",
                probability=0.9369364,
                label="Crude & Condensates",
                source="model",
            )

            assert expected == deserialized

    def test_instantiate_from_dict(self):
        dictionary = {
            "id": "b68cbb746f8b9098c50e2ba36bcad83001a53bd362e9031fb49085d02c36659c",
            "layer": "group",
            "probability": 0.4756425,
            "source": "model",
            "label": "Clean products",
        }

        p = ProductEntity(**dictionary)

        assert p.source == "model"



from dataclasses import dataclass

from vortexasdk.api import ISODate
from vortexasdk.api.serdes import FromDictMixin


@dataclass(frozen=True)
class TimeSeriesItem(FromDictMixin):
    """
    Generic container class holding a _key_, _value_ pair, and a _count_ of records contributing to the given value.

    For example, this class could hold the total tonnage exported (value) on 2019-01-01 (key), and the count of cargo
    movements contributing to this tonnage aggregate, ie the number of cargo movements on this day (count).
    """

    key: ISODate
    value: float
    count: int



from typing import List

import pandas as pd

from vortexasdk.api import Corporation
from vortexasdk.api.search_result import Result
from vortexasdk.logger import get_logger
from vortexasdk.result_conversions import create_dataframe, create_list

logger = get_logger(__name__)


class CorporationsResult(Result):
    """Container class that holds the result obtained from calling the `Vessels` endpoint."""

    def to_list(self) -> List[Corporation]:
        """Represent vessels as a list."""
        # noinspection PyTypeChecker
        return create_list(super().to_list(), Corporation)

    def to_df(self, columns=None) -> pd.DataFrame:
        """
        Represent corporations as a `pd.DataFrame`.

        # Arguments
            columns: The corporation features we want in the dataframe. Enter `columns='all'` to include all features.
            Defaults to `columns = ['id', 'name', 'corporate_entity_type']`.


        # Returns
        `pd.DataFrame` of corporations.

        """
        return create_dataframe(
            columns=columns,
            default_columns=DEFAULT_COLUMNS,
            data=super().to_list(),
            logger_description="Corporations",
        )


DEFAULT_COLUMNS = ["id", "name", "corporate_entity_type"]



import os

import requests
from vortexasdk.endpoints.endpoints import API_FQDN

all_tests_pass = True


def check_api_key_present():
    global all_tests_pass
    api_key = os.getenv("VORTEXA_API_KEY")
    if api_key is None:
        print("Environment variable VORTEXA_API_KEY is not set.")
        print(
            "         Please set this environment variable, this is the recommended way to authenticate with the SDK."
            "\n         You may need to restart your python interpreter / bash terminal for the API Key to be ingested."
            "\n         Note: The SDK will work without an environment variable, you'll be prompted to interactively enter your password."
        )
    else:
        print("Environment variable VORTEXA_API_KEY is set correctly.")


def check_can_connect_to_vortexa_api():
    global all_tests_pass

    url = f"{API_FQDN}/health-check"
    try:
        response = requests.get(url)
        reason, status_code = response.reason, response.status_code
    except Exception:
        reason, status_code = None, None

    if status_code == 200:
        print(f"Python successfully connected to {url}")
    else:
        all_tests_pass = False
        print(f"Python unable to connect to {url}")
        print(f"         status code: {status_code}")
        print(f"         reason: {reason}")
        print(f"         Check your internet connectivity / VPN settings.")


def check_can_import_vortexasdk():
    global all_tests_pass

    try:
        import vortexasdk

        print("Python successfully imported vortexasdk")
    except ImportError:
        all_tests_pass = False
        print("Python unable to import vortexasdk")


def check_can_retrieve_geographies():
    global all_tests_pass

    # noinspection PyBroadException
    try:
        from vortexasdk import Geographies

        europe = (
            "f39d455f5d38907394d6da3a91da4e391f9a34bd6a17e826d6042761067e88f4"
        )
        geography = Geographies().reference(europe)
        assert geography["id"] == europe
        print(
            "Python successfully retrieved a sample piece of reference data"
        )
    except Exception as e:
        all_tests_pass = False
        print(e)
        print("Python unable to retrieve a sample piece of reference data")


def run_all_checks():
    print("Running Vortexa SDK setup check")
    check_api_key_present()
    check_can_connect_to_vortexa_api()
    check_can_import_vortexasdk()
    check_can_retrieve_geographies()

    if all_tests_pass:
        print("All tests passed, the SDK is correctly setup!")
    else:
        print("One or more tests have failed.")
    print()


if __name__ == "__main__":
    run_all_checks()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List

import pandas as pd


@dataclass
class Result(ABC):
    """Abstract Container that holds a list of *_records*."""

    _records: List

    def to_list(self) -> List:
        """Represent *_records* as a list."""
        return self._records

    @abstractmethod
    def to_df(self, columns=None) -> pd.DataFrame:
        """Represent *_records* as a `pd.DataFrame` with given columns."""
        pass

    def __len__(self):
        """Delegate to *_records*."""
        return len(self._records)

    def __str__(self):
        """Delegate to *_records*."""
        return str(self._records)

    def __iter__(self):
        """Delegate to *_records*."""
        return iter(self._records)

    def __getitem__(self, item):
        """Delegate to *_records*."""
        return self._records.__getitem__(item)



from tests.testcases import TestCaseUsingRealAPI
from docs.utils import to_markdown
from vortexasdk import Geographies


class TestGeographiesReal(TestCaseUsingRealAPI):
    def test_search(self):
        geographies = Geographies().search(term=["Liverpool", "Southampton"])
        names = [g["name"] for g in geographies]

        assert "Liverpool [GB]" in names

    def test_search_to_df(self):
        geographies = (
            Geographies().search(term=["Liverpool", "Southampton"]).to_df()
        )

        print(to_markdown(geographies))



"""Conversions between various entites."""
from vortexasdk.conversions.corporations import convert_to_corporation_ids
from vortexasdk.conversions.geographies import convert_to_geography_ids
from vortexasdk.conversions.products import convert_to_product_ids
from vortexasdk.conversions.vessels import convert_to_vessel_ids



from unittest import TestCase

import pandas as pd

from tests.mock_client import example_corporations
from vortexasdk.endpoints.corporations_result import CorporationsResult


class TestCorporationsSearchResult(TestCase):
    cr = CorporationsResult(example_corporations)

    def test_to_df(self):
        df = self.cr.to_df(["name", "corporate_entity_type"])

        expected = pd.DataFrame(
            {
                "name": {0: "3J", 1: "5XJAPANESE"},
                "corporate_entity_type": {
                    0: ["commercial_owner"],
                    1: ["commercial_owner"],
                },
            }
        )
        assert expected.equals(df)

    def test_to_list(self):
        names = [x.name for x in self.cr.to_list()]

        assert names == ["3J", "5XJAPANESE"]



from tests.testcases import TestCaseUsingRealAPI
from datetime import datetime
from tests.timer import Timer
from vortexasdk.endpoints.eia_forecasts import EIAForecasts
from docs.utils import to_markdown


class TestEIAForecastsReal(TestCaseUsingRealAPI):
    def test_search_preset_crude_imports(self):

        preset = "padd1-crude-imports"
        filter_time_min = datetime(2020, 1, 20)
        filter_time_max = datetime(2020, 1, 24)

        forecasts = EIAForecasts().search(
            preset=preset,
            filter_time_min=filter_time_min,
            filter_time_max=filter_time_max,
        )
        assert len(forecasts) == 1
        values = [g.date for g in forecasts.to_list()]

        assert "2020-01-24T00:00:00.000Z" in values

        print(to_markdown(forecasts.to_df()))

    def test_search_preset_gasoline_exports(self):

        preset = "us-gasoline-exports"
        filter_time_min = datetime(2020, 3, 20)
        filter_time_max = datetime(2020, 4, 20)

        forecasts = EIAForecasts().search(
            preset=preset,
            filter_time_min=filter_time_min,
            filter_time_max=filter_time_max,
        )
        assert len(forecasts) == 5

        values = [g.date for g in forecasts.to_list()]
        assert "2020-04-17T00:00:00.000Z" in values
        assert "2020-04-10T00:00:00.000Z" in values
        assert "2020-04-03T00:00:00.000Z" in values
        assert "2020-03-27T00:00:00.000Z" in values
        assert "2020-03-20T00:00:00.000Z" in values

        print(to_markdown(forecasts.to_df()))



from dataclasses import dataclass
from typing import Optional

from vortexasdk.api.serdes import FromDictMixin


@dataclass(frozen=True)
class EIAForecast(FromDictMixin):
    """Represent a EIA forecast record returned by the API."""

    date: str
    forecast_fri: float
    value: Optional[int]
    stocks: Optional[int]
    cover: Optional[float]
    runs: Optional[float]



from tests.testcases import TestCaseUsingRealAPI
from vortexasdk.conversions import convert_to_vessel_ids
from vortexasdk.endpoints.vessels import Vessels


class TestConvert(TestCaseUsingRealAPI):
    def test_convert_name_to_vessel_ids(self):
        result = convert_to_vessel_ids(["OCEAN"])

        assert len(result) >= 10

    def test_convert_imo_to_vessel_ids(self):
        result = convert_to_vessel_ids([9083287])
        assert result == [
            "84a82843ec84ac6d67b65c50056eff78e0d58e6b9fc7a5ba9adc6c0442162cf4"
        ]


    def test_convert_class_to_vessel_ids(self):
        result = convert_to_vessel_ids(["panamax"])

        assert result == [
            v.id for v in Vessels().search(vessel_classes="panamax").to_list()
        ]

    def test_convert_combination_to_vessel_ids(self):
        from_mmsi = convert_to_vessel_ids([563064200])
        from_imo = convert_to_vessel_ids([9083287])
        from_name = convert_to_vessel_ids(["OCEAN"])
        from_class = convert_to_vessel_ids(["aframax"])

        result = set(
            convert_to_vessel_ids([563064200, 9083287, "OCEAN", "aframax"])
        )
        assert result == set(from_mmsi + from_imo + from_name + from_class)



from typing import List

from vortexasdk.api import ID
from vortexasdk.api.id import split_ids_other
from vortexasdk.api.shared_types import IDsNames
from vortexasdk.logger import get_logger
from vortexasdk.operations import Search
from vortexasdk.utils import convert_to_list

logger = get_logger(__name__)


def _convert_to_ids(ids_or_names: IDsNames, searcher: Search) -> List[ID]:
    """Convert containing a mix of IDs and names to a list of IDs."""
    ids_or_names_list = convert_to_list(ids_or_names)

    ids, others = split_ids_other(ids_or_names_list)
    if len(others) == 0:
        return ids
    else:
        return ids + _search_ids(searcher, term=others)


def _search_ids(searcher: Search, **kwargs) -> List[ID]:
    """Find IDs matching a given list of search terms."""
    logger.info(
        f"Searching {searcher.__class__.__name__} matching search terms: {kwargs}"
    )

    results = searcher.search(**kwargs)

    id_to_name = {r["id"]: r["name"] for r in results}

    logger.info(
        f"Found {len(results)} {searcher.__class__.__name__}: {id_to_name}"
    )

    return list(id_to_name.keys())



from dataclasses import dataclass
from typing import List

from vortexasdk.api.serdes import FromDictMixin
from vortexasdk.api.shared_types import Entity, EntityWithProbability, Node


@dataclass(frozen=True)
class Product(Node, FromDictMixin):
    """
    Represent a Product reference record returned by the API.

    [Product Further Documentation](https://docs.vortexa.com/reference/GET/reference/products/%7Bid%7D)
    """

    layer: List[str]
    meta: dict
    hierarchy: List[Entity]


@dataclass(frozen=True)
class ProductEntity(EntityWithProbability):
    """

    Represents a single product layer of a hierarchical product tree.

    [Further Documentation](https://docs.vortexa.com/reference/intro-product-entities)
    """



from abc import ABC
from typing import List

from vortexasdk.api import ID


class AbstractVortexaClient(ABC):
    """Base client."""

    def get_reference(self, resource: str, id: ID) -> str:
        """Lookup reference data."""
        raise NotImplementedError

    def search(self, resource: str, **data) -> List:
        """Search."""
        raise NotImplementedError



from unittest import TestCase

import jsons

from vortexasdk.api.corporation import CorporateEntity


class TestCorporateEntity(TestCase):
    def test_serialize(self):
        with open("tests/api/examples/corporate_entity1.json", "r") as f:
            serialized = f.read()
            deserialized = jsons.loads(serialized, CorporateEntity)

            expected = CorporateEntity(
                id="cbd7dfe8a9fb0fa0ce3252ce7643437db6a32d0947a0c23d68dc5dea2f2d65d7",
                layer="commercial_owner",
                probability=1,
                label="NGM Energy",
                source="external",
            )

            assert expected == deserialized



from typing import List

import pandas as pd

from vortexasdk.api.search_result import Result
from vortexasdk.api.timeseries_item import TimeSeriesItem
from vortexasdk.logger import get_logger
from vortexasdk.result_conversions import create_dataframe, create_list

logger = get_logger(__name__)


class TimeSeriesResult(Result):
    """Container class that holds the result obtained from calling a time series endpoint."""

    def to_list(self) -> List[TimeSeriesItem]:
        """Represents time series as a list."""
        # noinspection PyTypeChecker
        return create_list(super().to_list(), TimeSeriesItem)

    def to_df(self, columns=None) -> pd.DataFrame:
        """Represents the timeseries as a dataframe.

        Returns a `pd.DataFrame`, of time series items with columns:
         key: The time series key
         value: The value of the time series for a given key
         count: The number of records contributing to this time series record.

        # Example:

        If we're aggregating Crude exports in tonnes by day, then the `key` column holds the date,
        the `value` column holds the Crude exports on that day, and the `count` column holds
        the number of cargo movements contributing towards this day's tonnage.

        """
        df = create_dataframe(
            columns=columns,
            default_columns=DEFAULT_COLUMNS,
            data=super().to_list(),
            logger_description="TimeSeries",
        )

        df["key"] = pd.to_datetime(df["key"])

        return df


DEFAULT_COLUMNS = ["key", "value", "count"]



from tests.testcases import TestCaseUsingRealAPI
from vortexasdk.endpoints.products import Products
from vortexasdk.endpoints.products_result import DEFAULT_COLUMNS


class TestProductsReal(TestCaseUsingRealAPI):
    def test_search(self):
        products = Products().search().to_list()
        assert len(products) > 0

        print([x.name for x in products[:5]])

    def test_search_dataframe(self):
        df = Products().search().to_df()
        assert list(df.columns) == DEFAULT_COLUMNS
        assert len(df) > 0

    def test_search_dataframe_subset_of_cols(self):
        df = Products().search().to_df(columns=["name", "id"])
        assert list(df.columns) == ["name", "id"]
        assert len(df) > 0

    def test_search_exact_match(self):
        search_term = "Gasoil"

        # Confirm that searching our search term does indeed yield more than one result
        products = Products().search(search_term).to_list()
        if len(products) <= 1:
            raise Exception(
                f"Unable to perform exact match test on {search_term}."
            )

        products_exact = (
            Products().search(search_term, exact_term_match=True).to_list()
        )

        assert {p.name for p in products_exact} == {search_term}

    def test_load_all(self):
        all_products = Products().load_all()

        assert len(all_products) > 20

    def test_search_multiple_terms_to_dataframe(self):
        df = (
            Products()
            .search(
                term=["diesel", "fuel oil", "grane", "lng", "lpg", "crude"]
            )
            .to_df("all")
        )

        expected_columns = {
            "id",
            "name",
            "layer.0",
            "leaf",
            "parent.0.name",
            "parent.0.layer.0",
            "parent.0.id",
            "meta.api_min",
            "meta.api_max",
            "ref_type",
            "meta.sulphur_min",
            "meta.sulphur_max",
        }

        assert expected_columns.issubset(set(df.columns))



from tests.testcases import TestCaseUsingRealAPI
from docs.utils import to_markdown
from vortexasdk import Corporations


class TestCorporationsReal(TestCaseUsingRealAPI):
    def test_search(self):
        df = Corporations().search().to_df()

        print(to_markdown(df.head(2)))

    def test_load_all(self):
        all_corporations = Corporations().load_all()

        assert len(all_corporations) > 1000



from typing import List, Union

from vortexasdk.endpoints.geographies import Geographies
from vortexasdk.api import ID
from vortexasdk.conversions.conversions import _convert_to_ids


def convert_to_geography_ids(
    ids_or_names_list: List[Union[ID, str]]
) -> List[ID]:
    """
    Convert a mixed list of names or IDs to geography ids.

    # Example
    ```
    >>> convert_to_geography_ids(["Rotterdam [NL]"])
    [...]

    ```

    # Example
    ```
    >>> convert_to_geography_ids(["Rotterdam [NL]", "b514a3bfd0b87d91984f43d9ee5071fb3a063ec309f2fe486ca3d2e58b61d683"])
    [...]

    ```

    """
    return _convert_to_ids(ids_or_names_list, Geographies())



from tests.testcases import TestCaseUsingRealAPI
from tests.timer import Timer
from vortexasdk.endpoints.attributes import Attributes


class TestAttributesReal(TestCaseUsingRealAPI):
    def test_search_ids(self):
        ids = [
            "14c7b073809eb565",
            "3ace0e050724707b",
        ]

        attributes = Attributes().search(ids=ids).to_list()
        assert len(attributes) == 2

        print([x.name for x in attributes])

    def test_search_filters_term(self):
        term = ["Open Loop Hybrid Ready"]

        attributes = Attributes().search(term=term).to_list()

        actual = {x.name for x in attributes}

        assert actual == set(term)

    def test_search_ids_dataframe(self):
        ids = [
            "4729c38b99262b73",
            "478fca39000c49d6",
        ]

        df = Attributes().search(ids=ids).to_df()
        assert list(df.columns) == ["id", "name", "type"]
        assert len(df) == 2

    def test_load_all(self):
        all_attributes = Attributes().load_all()

        assert len(all_attributes) > 40

    def test_search_load_all_attributes(self):
        with Timer("Search"):
            result = Attributes().search()

        with Timer("Serialize"):
            result.to_list()

        with Timer("Dataframe"):
            df = result.to_df()
            print(df.head())

        assert len(result) >= 40

    def test_search_is_case_agnostic(self):
        uppercase = {v.id for v in Attributes().search(term="open").to_list()}
        lowercase = {v.id for v in Attributes().search(term="OpEn").to_list()}

        assert uppercase == lowercase



from docs.utils import to_markdown
from tests.testcases import TestCaseUsingRealAPI
from vortexasdk import Geographies


class TestGeographiesReal(TestCaseUsingRealAPI):
    def test_load_all(self):
        all_geogs = Geographies().load_all()

        assert len(all_geogs) > 1000

    def test_search_empty_args(self):
        Geographies().search()

    def test_search_to_df(self):
        geographies = (
            Geographies().search(term=["Liverpool", "Southampton"]).to_df()
        )

        print(to_markdown(geographies))

    def test_search_to_list(self):
        geographies = (
            Geographies().search(term=["Liverpool", "Southampton"]).to_list()
        )
        names = [g.name for g in geographies]
        assert "Liverpool [GB]" in names



from tests.testcases import TestCaseUsingRealAPI
from vortexasdk import Products, Geographies, Corporations, Vessels

endpoints_and_searchterms = [
    (Products(), "Gasoil"),
    (Geographies(), "China"),
    (Corporations(), "Oil"),
    (Vessels(), "Ocean"),
]


class TestSearchReal(TestCaseUsingRealAPI):
    def test_search_exact_match_yields_fewer_results_than_non_exact_match(
        self,
    ):

        for endpoint, term in endpoints_and_searchterms:
            result_loose_match = endpoint.search(
                term=term, exact_term_match=False
            )
            result_exact_match = endpoint.search(
                term=term, exact_term_match=True
            )

            assert len(result_exact_match) < len(result_loose_match)

    def test_search_exact_match_yields_exact_matches_only(self):

        for endpoint, term in endpoints_and_searchterms:
            result_exact_match = endpoint.search(
                term=term, exact_term_match=True
            )

            actual_names = {e["name"] for e in result_exact_match}

            # result must be the exact term, or must contain no results if there's no reference objects with that names
            assert actual_names == {term} or actual_names == set()



from requests import Response, Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry

_HEADERS = {"Content-Type": "application/json"}


# Inspired by https://www.peterbe.com/plog/best-practice-with-retries-with-requests
def _requests_retry_session(
    retries=6,
    backoff_factor=1,
    status_forcelist=(500, 502, 504),
    session=None,
) -> Session:
    """Instantiate a session with Retry backoff."""
    session = session or Session()

    retry = Retry(
        raise_on_redirect=False,
        raise_on_status=False,
        method_whitelist=["POST", "GET"],
        status=retries,
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)

    return session


def retry_get(*args, **kwargs) -> Response:
    with _requests_retry_session() as s:
        return s.get(headers=_HEADERS, *args, **kwargs)


def retry_post(*args, **kwargs) -> Response:
    with _requests_retry_session() as s:
        return s.post(headers=_HEADERS, *args, **kwargs)



from unittest import TestCase

from vortexasdk.utils import (
    convert_to_list,
    convert_values_to_list,
    filter_exact_match,
)


class TestUtils(TestCase):
    def test_single_item_list(self):
        self.assertEqual(convert_to_list("a"), ["a"])

    def test_multiple_items_list(self):
        self.assertCountEqual(convert_to_list(["a", "b"]), ["a", "b"])

    def test_convert_values_to_list(self):
        actual = convert_values_to_list({1: None, 2: "b", 3: ["c"]})

        expected = {1: [], 2: ["b"], 3: ["c"]}

        assert actual == expected

    def test_filter_doesn_not_match_part_name(self):
        search_result = [{"name": "China"}, {"name": "South China"}]

        self.assertEqual(filter_exact_match("Chi", search_result), [])

    def test_filter_exact_match_single_name(self):
        search_result = [{"name": "China"}, {"name": "South China"}]

        actual = filter_exact_match("China", search_result)

        expected = [{"name": "China"}]

        self.assertEqual(actual, expected)

    def test_filter_exact_match_multiple_names(self):
        search_result = [
            {"name": "China"},
            {"name": "South China"},
            {"name": "United States"},
            {"name": "United Arab Emirates"},
        ]

        actual = filter_exact_match(["China", "United States"], search_result)
        expected = [{"name": "China"}, {"name": "United States"}]

        self.assertCountEqual(actual, expected)



from dataclasses import dataclass

from vortexasdk.api.serdes import FromDictMixin
from vortexasdk.api.shared_types import Node


@dataclass(frozen=True,)
class Attribute(Node, FromDictMixin):
    """
    Represent an Attribute reference record returned by the API.
    """

    id: str
    name: str
    type: str



from unittest import TestCase

import jsons
from typing import List

from vortexasdk.api import (
    CargoEvent,
    CargoMovement,
    CorporateEntity,
    GeographyEntity,
    ProductEntity,
    VesselEntity,
    ParentID
)
from vortexasdk.api.entity_flattening import (
    convert_cargo_movement_to_flat_dict,
)
from vortexasdk.api.serdes import serialize_to_dict


class TestCargoMovement(TestCase):
    dictionary = {
        "cargo_movement_id": "00886b05a0747522b67322f50123ee60e61e219fc9a9c6011be1a1dade65f63e",
        "quantity": 4401,
        "status": "unloaded_state",
        "vessels": [
            VesselEntity(
                **{
                    "id": "9cbf7c0fc6ccf1dfeacde02b87f3b6b1653030f560d4fc677bf1d7d0be8f8449",
                    "mmsi": 255804460,
                    "imo": 9480980,
                    "name": "JOHANN ESSBERGER",
                    "dwt": 5260,
                    "cubic_capacity": 6100,
                    "vessel_class": "tiny_tanker",
                    "corporate_entities": [
                        CorporateEntity(
                            **{
                                "id": "f9bd45e65e292909a7b751b0026dcf7795c6194b3c0712910a241caee32c99b8",
                                "label": "Essberger J.T.",
                                "layer": "commercial_owner",
                                "probability": 1,
                                "source": "external",
                            }
                        )
                    ],
                    "start_timestamp": "2019-10-18T21:38:34+0000",
                    "end_timestamp": "2019-10-25T00:40:46+0000",
                    "fixture_fulfilled": False,
                    "voyage_id": "401f0e74fc42401248a484aca2e9955dea885378796f7f4d0bc8e92c35ea270a",
                    "tags": [],
                    "status": "vessel_status_laden_known",
                }
            )
        ],
        "product": [
            ProductEntity(
                **{
                    "id": "b68cbb746f8b9098c50e2ba36bcad83001a53bd362e9031fb49085d02c36659c",
                    "layer": "group",
                    "probability": 0.4756425,
                    "source": "model",
                    "label": "Clean products",
                }
            ),
            ProductEntity(
                **{
                    "id": "a75fcc09bfc7d16496de3336551bc52b5891838bb7c22356d2cb65451587d1e5",
                    "layer": "group_product",
                    "probability": 0.4756425,
                    "source": "model",
                    "label": "Biodiesel",
                }
            ),
            ProductEntity(
                **{
                    "id": "9d52ede1cff0421a8cd7283b0171afe8d23f519dca5f4e489734522f9cdf804c",
                    "layer": "grade",
                    "probability": 0.4756425,
                    "source": "model",
                    "label": "Biodiesel Feedstock",
                }
            ),
        ],
        "events": [
            CargoEvent(
                **{
                    "event_type": "cargo_port_load_event",
                    "location": [
                        GeographyEntity(
                            **{
                                "id": "2dfc3d43a21697c02ec3b2700b3b570a6ed1abb66d01c4fe6310ef362fcf6653",
                                "layer": "country",
                                "label": "Netherlands",
                                "source": "model",
                                "probability": 1,
                            }
                        )
                    ],
                    "probability": 1,
                    "pos": [4.29914090037834, 51.87936163368058],
                    "start_timestamp": "2019-10-18T21:38:34+0000",
                    "end_timestamp": "2019-10-20T16:41:49+0000",
                }
            )
        ],
        "parent_ids": [
            ParentID(
                **{
                    "id": "9d52ede1cff0421a8cd7283b0171afe8d23f519dca5f4e489734522f9cdf804c",
                    "splinter_timestamp": "2019-10-20T16:41:49+0000",
                }
            )
        ]
    }

    cm = CargoMovement(**dictionary)

    def test_serialize(self):
        with open("tests/api/examples/cargo_movements.json", "r") as f:
            serialized = f.read()
            deserialized = jsons.loads(serialized, List[CargoMovement])

            assert [self.cm] == deserialized

    def test_convert_to_flat_dict(self):
        flat = convert_cargo_movement_to_flat_dict(serialize_to_dict(self.cm))

        expected = {
            "cargo_movement_id": "00886b05a0747522b67322f50123ee60e61e219fc9a9c6011be1a1dade65f63e",
            "events.cargo_port_load_event.0.end_timestamp": "2019-10-20T16:41:49+0000",
            "events.cargo_port_load_event.0.event_type": "cargo_port_load_event",
            "events.cargo_port_load_event.0.location.country.id": "2dfc3d43a21697c02ec3b2700b3b570a6ed1abb66d01c4fe6310ef362fcf6653",
            "events.cargo_port_load_event.0.location.country.label": "Netherlands",
            "events.cargo_port_load_event.0.location.country.layer": "country",
            "events.cargo_port_load_event.0.location.country.probability": 1,
            "events.cargo_port_load_event.0.location.country.source": "model",
            "events.cargo_port_load_event.0.pos.0": 4.29914090037834,
            "events.cargo_port_load_event.0.pos.1": 51.87936163368058,
            "events.cargo_port_load_event.0.probability": 1,
            "events.cargo_port_load_event.0.start_timestamp": "2019-10-18T21:38:34+0000",
            "events.cargo_port_load_event.0.vessel_id": None,
            "product.group.id": "b68cbb746f8b9098c50e2ba36bcad83001a53bd362e9031fb49085d02c36659c",
            "product.group.label": "Clean products",
            "product.group.layer": "group",
            "product.group.probability": 0.4756425,
            "product.group.source": "model",
            "product.group_product.id": "a75fcc09bfc7d16496de3336551bc52b5891838bb7c22356d2cb65451587d1e5",
            "product.group_product.label": "Biodiesel",
            "product.group_product.layer": "group_product",
            "product.group_product.probability": 0.4756425,
            "product.group_product.source": "model",
            "product.grade.id": "9d52ede1cff0421a8cd7283b0171afe8d23f519dca5f4e489734522f9cdf804c",
            "product.grade.label": "Biodiesel Feedstock",
            "product.grade.layer": "grade",
            "product.grade.probability": 0.4756425,
            "product.grade.source": "model",
            "quantity": 4401,
            "status": "unloaded_state",
            "vessels.0.corporate_entities.commercial_owner.id": "f9bd45e65e292909a7b751b0026dcf7795c6194b3c0712910a241caee32c99b8",
            "vessels.0.corporate_entities.commercial_owner.label": "Essberger J.T.",
            "vessels.0.corporate_entities.commercial_owner.layer": "commercial_owner",
            "vessels.0.corporate_entities.commercial_owner.probability": 1,
            "vessels.0.corporate_entities.commercial_owner.source": "external",
            "vessels.0.cubic_capacity": 6100,
            "vessels.0.dwt": 5260,
            "vessels.0.end_timestamp": "2019-10-25T00:40:46+0000",
            "vessels.0.fixture_fulfilled": False,
            "vessels.0.fixture_id": None,
            "vessels.0.id": "9cbf7c0fc6ccf1dfeacde02b87f3b6b1653030f560d4fc677bf1d7d0be8f8449",
            "vessels.0.imo": 9480980,
            "vessels.0.mmsi": 255804460,
            "vessels.0.name": "JOHANN ESSBERGER",
            "vessels.0.year": None,
            "vessels.0.scrubber": None,
            "vessels.0.flag": None,
            "vessels.0.ice_class": None,
            "vessels.0.propulsion": None,
            "vessels.0.start_timestamp": "2019-10-18T21:38:34+0000",
            "vessels.0.status": "vessel_status_laden_known",
            "vessels.0.vessel_class": "tiny_tanker",
            "vessels.0.voyage_id": "401f0e74fc42401248a484aca2e9955dea885378796f7f4d0bc8e92c35ea270a",
            "parent_ids.0.id": "9d52ede1cff0421a8cd7283b0171afe8d23f519dca5f4e489734522f9cdf804c",
            "parent_ids.0.splinter_timestamp": "2019-10-20T16:41:49+0000"
        }

        assert flat == expected



"""Vortexa endpoints."""
from vortexasdk.endpoints.attributes import Attributes
from vortexasdk.endpoints.cargo_movements import CargoMovements
from vortexasdk.endpoints.cargo_timeseries import CargoTimeSeries
from vortexasdk.endpoints.vessel_movements import VesselMovements
from vortexasdk.endpoints.corporations import Corporations
from vortexasdk.endpoints.geographies import Geographies
from vortexasdk.endpoints.products import Products
from vortexasdk.endpoints.vessels import Vessels
from vortexasdk.endpoints.eia_forecasts import EIAForecasts



"""Vessels Endpoint."""
from typing import Dict, List, Union

from vortexasdk.api.id import ID
from vortexasdk.endpoints.endpoints import VESSELS_REFERENCE
from vortexasdk.endpoints.vessels_result import VesselsResult
from vortexasdk.operations import Reference, Search
from vortexasdk.utils import convert_to_list


class Vessels(Reference, Search):
    """Vessels endpoint."""

    def __init__(self):
        """Instantiate endpoint using reference endpoint."""
        Reference.__init__(self, VESSELS_REFERENCE)
        Search.__init__(self, VESSELS_REFERENCE)

    def load_all(self) -> VesselsResult:
        """Load all vessels."""
        return self.search()

    def search(
        self,
        term: Union[str, List[str]] = None,
        ids: Union[str, List[str]] = None,
        vessel_classes: Union[str, List[str]] = None,
        vessel_product_types: Union[ID, List[ID]] = None,
        vessel_scrubbers: str = "disabled",
        exact_term_match: bool = False,
    ) -> VesselsResult:
        """
        Find all vessels matching given search arguments. Search arguments are combined in an AND manner.

        # Arguments
            term: The name(s) (or partial name(s)) of a vessel we'd like to search

            ids: ID or IDs of vessels we'd like to search

            vessel_classes: vessel_class (or list of vessel classes) we'd like to search. Each vessel class must be one of `"tiny_tanker" , "general_purpose" , "handysize" , "handymax" , "panamax", "aframax" , "suezmax" , "vlcc_plus" , "sgc" , "mgc" , "lgc" , "vlgc"`. Refer to [VortexaAPI Vessel Entities](https://docs.vortexa.com/reference/intro-vessel-entities) for the most up-to-date list of vessel classes.

            vessel_product_types: A product ID, or list of product IDs to filter on, searching vessels _currently_ carrying these products.

            vessel_scrubbers: An optional filter to filter on vessels with or without scrubbers.
             To disable the filter (the default behaviour), enter 'disabled'.
             To only include vessels with scrubbers, enter 'inc'.
             To exclude vessels with scrubbers, enter 'exc'.

             exact_term_match: Search on only exact term matches, or allow similar matches.
                 e.g. When searching for "Ocean" with `exact_term_match=False`, then the SDK will yield vessels named
                ['Ocean', 'Ocean Wisdom', ...] etc. When `exact_term_match=True`,
                the SDK will only yield the vessel named `Ocean`.


        # Returns
        List of vessels matching the search arguments.


        # Examples

        - Let's find all the VLCCs with 'ocean' in their name, or related names.

        ```python
        >>> from vortexasdk import Vessels
        >>> vessels_df = Vessels().search(vessel_classes='vlcc', term='ocean').to_df(columns=['name', 'imo', 'mmsi', 'related_names'])

        ```
        |    | name         |     imo |      mmsi | related_names             |
        |---:|:-------------|--------:|----------:|:--------------------------|
        |  0 | OCEANIS      | 9532757 | 241089000 | ['OCEANIS']               |
        |  1 | AEGEAN       | 9732553 | 205761000 | ['GENER8 OCEANUS']        |
        |  2 | OCEANIA      | 9246633 | 205753000 | ['OCEANIA'| 'TI OCEANIA'] |
        |  3 | ENEOS OCEAN  | 9662875 | 432986000 | ['ENEOS OCEAN']           |
        |  4 | OCEAN LILY   | 9284960 | 477178100 | ['OCEAN LILY']            |
        |  5 | SHINYO OCEAN | 9197868 | 636019316 | ['SHINYO OCEAN']          |
        |  6 | NASHA        | 9079107 | 370497000 | ['OCEANIC']               |
        |  7 | HUMANITY     | 9180281 | 422204700 | ['OCEAN NYMPH']           |

        Note the `term` search also looks for vessels with matching `related_names`


        - Let's find all the vessels currently carrying Crude.

        ```python
        >>> from vortexasdk import Vessels, Products
        >>> crude = [p.id for p in Products().search(term="crude").to_list() if 'group' in p.layer]
        >>> vessels_df = Vessels().search(vessel_product_types=crude).to_df()

        ```

        # Further Documentation

        [VortexaAPI Vessel Reference](https://docs.vortexa.com/reference/POST/reference/vessels)

        """
        api_params = {
            "term": [str(e) for e in convert_to_list(term)],
            "ids": convert_to_list(ids),
            "vessel_product_types": convert_to_list(vessel_product_types),
            "vessel_classes": [
                v.lower() for v in convert_to_list(vessel_classes)
            ],
            "vessel_scrubbers": vessel_scrubbers,
        }

        return VesselsResult(
            super().search(exact_term_match=exact_term_match, **api_params)
        )

    def reference(self, id: ID) -> Dict:
        """
        Perform a vessel lookup.

        # Arguments
            id: Vessel ID to lookup

        # Returns
        Vessel record matching the ID

        # Further Documentation:
        [VortexaAPI Vessel Reference](https://docs.vortexa.com/reference/GET/reference/vessels/%7Bid%7D)

        """
        return super().reference(id)


AVAILABLE_VESSEL_CLASSES = [
    "tiny_tanker",
    "general_purpose",
    "handysize",
    "handymax",
    "panamax",
    "aframax",
    "suezmax",
    "vlcc_plus",
    "vlcc",
    "sgc",
    "mgc",
    "lgc",
    "vlgc",
]



from unittest import TestCase, skipIf

from tests.config import SKIP_TAGS
from tests.mock_client import MockVortexaClient
from vortexasdk.client import create_client, set_client


@skipIf("real" in SKIP_TAGS, "Skipping tests that hit the real API server.")
class TestCaseUsingRealAPI(TestCase):
    def setUp(self) -> None:
        set_client(create_client())


class TestCaseUsingMockAPI(TestCase):
    def setUp(self) -> None:
        set_client(MockVortexaClient())



import functools
import os
from multiprocessing.pool import Pool
from typing import List

import pandas as pd

from vortexasdk.api import CargoMovement
from vortexasdk.api.entity_flattening import (
    convert_cargo_movement_to_flat_dict,
)
from vortexasdk.api.search_result import Result
from vortexasdk.result_conversions import create_dataframe, create_list
from vortexasdk.logger import get_logger

logger = get_logger(__name__)


class CargoMovementsResult(Result):
    """
    Container class holdings search results returns from the cargo movements endpoint.

    This class has two methods, `to_list()`, and `to_df()`, allowing search results to be represented as a list of `CargoMovements`,
     or as a `pd.DataFrame` , respectively.
    """

    def to_list(self) -> List[CargoMovement]:
        """Represent cargo movements as a list of `CargoMovementEntity`s."""
        # noinspection PyTypeChecker
        return create_list(super().to_list(), CargoMovement)

    def to_df(self, columns=None) -> pd.DataFrame:
        """
        Represent cargo movements as a `pd.DataFrame`.

        # Arguments
            columns: Output columns present in the `pd.DataFrame`.
            Enter `columns='all'` to return all available columns.
            Enter `columns=None` to use `cargo_movements.DEFAULT_COLUMNS`.

        # Returns
        `pd.DataFrame`, one row per cargo movement.


        ## Notes

        A cargo movement is a complicated, nested structure. Between it's point of loading and discharge, a cargo
         movement may be carried by _N_ or more vessels, with _N-1_ associated STS events. Each of these _N_ vessels
         could have an associated corporate owner, charterer, time charterer... etc.

         In order to represent a cargo movement as a flat (not nested) record in a dataframe, the sdk flattens the cargo
         movement, generating many columns in the process.

         The columns are logically named. Let's say that a cargo is transferred between 4 vessels en route from a load
         in Rotterdam to a discharge in New York. This is represented as 1 `cargo_port_unload_event`, followed by
          3 `cargo_sts_event`s, and finally 1 `cargo_port_unload_event`.

         In this example the name of the 1st vessel, is found in the `vessels.0.name` column (we're using zero-based
         numbering indexes). Likewise, the imo of the second vessel is found in the `vessels.1.imo` column.

         To find the name of the country in which the second STS event occured, we'd use the
         `events.cargo_sts_event.1.location.country.layer` column.

         Similarly, to find out when the first vessel started
         loading the cargo from Rotterdam, we'd use the `events.cargo_port_load_event.0.start_timestamp` column.

        By default, the columns returned are something along the lines of.
        ```python
        DEFAULT_COLUMNS = [
            'events.cargo_port_load_event.0.location.port.label',
            'events.cargo_port_unload_event.0.location.port.label',
            'product.group.label',
            'product.grade.label',
            'quantity',
            'vessels.0.name',
            'events.cargo_port_load_event.0.end_timestamp',
            'events.cargo_port_unload_event.0.start_timestamp',
        ]
        ```
        The exact default columns used can be found at `cargo_movements.DEFAULT_COLUMNS`

        A near complete list of columns is given below
        ```
        [
            'cargo_movement_id',
            'events.cargo_fso_load_event.0.end_timestamp',
            'events.cargo_fso_load_event.0.event_type',
            'events.cargo_fso_load_event.0.fso_vessel_id',
            'events.cargo_fso_load_event.0.fso_vessel_name',
            'events.cargo_fso_load_event.0.location.country.id',
            'events.cargo_fso_load_event.0.location.country.label',
            'events.cargo_fso_load_event.0.location.country.layer',
            'events.cargo_fso_load_event.0.location.country.probability',
            'events.cargo_fso_load_event.0.location.country.source',
            'events.cargo_fso_load_event.0.location.region.id',
            'events.cargo_fso_load_event.0.location.region.label',
            'events.cargo_fso_load_event.0.location.region.layer',
            'events.cargo_fso_load_event.0.location.region.probability',
            'events.cargo_fso_load_event.0.location.region.source',
            'events.cargo_fso_load_event.0.location.shipping_region.id',
            'events.cargo_fso_load_event.0.location.shipping_region.label',
            'events.cargo_fso_load_event.0.location.shipping_region.layer',
            'events.cargo_fso_load_event.0.location.shipping_region.probability',
            'events.cargo_fso_load_event.0.location.shipping_region.source',
            'events.cargo_fso_load_event.0.location.sts_zone.id',
            'events.cargo_fso_load_event.0.location.sts_zone.label',
            'events.cargo_fso_load_event.0.location.sts_zone.layer',
            'events.cargo_fso_load_event.0.location.sts_zone.probability',
            'events.cargo_fso_load_event.0.location.sts_zone.source',
            'events.cargo_fso_load_event.0.location.trading_block.id',
            'events.cargo_fso_load_event.0.location.trading_block.label',
            'events.cargo_fso_load_event.0.location.trading_block.layer',
            'events.cargo_fso_load_event.0.location.trading_block.probability',
            'events.cargo_fso_load_event.0.location.trading_block.source',
            'events.cargo_fso_load_event.0.location.trading_region.id',
            'events.cargo_fso_load_event.0.location.trading_region.label',
            'events.cargo_fso_load_event.0.location.trading_region.layer',
            'events.cargo_fso_load_event.0.location.trading_region.probability',
            'events.cargo_fso_load_event.0.location.trading_region.source',
            'events.cargo_fso_load_event.0.location.trading_subregion.id',
            'events.cargo_fso_load_event.0.location.trading_subregion.label',
            'events.cargo_fso_load_event.0.location.trading_subregion.layer',
            'events.cargo_fso_load_event.0.location.trading_subregion.probability',
            'events.cargo_fso_load_event.0.location.trading_subregion.source',
            'events.cargo_fso_load_event.0.pos.0',
            'events.cargo_fso_load_event.0.pos.1',
            'events.cargo_fso_load_event.0.probability',
            'events.cargo_fso_load_event.0.start_timestamp',
            'events.cargo_fso_load_event.0.to_vessel_id',
            'events.cargo_fso_load_event.0.to_vessel_name',
            'events.cargo_fso_unload_event.0.end_timestamp',
            'events.cargo_fso_unload_event.0.event_type',
            'events.cargo_fso_unload_event.0.from_vessel_id',
            'events.cargo_fso_unload_event.0.from_vessel_name',
            'events.cargo_fso_unload_event.0.fso_vessel_id',
            'events.cargo_fso_unload_event.0.fso_vessel_name',
            'events.cargo_fso_unload_event.0.location.country.id',
            'events.cargo_fso_unload_event.0.location.country.label',
            'events.cargo_fso_unload_event.0.location.country.layer',
            'events.cargo_fso_unload_event.0.location.country.probability',
            'events.cargo_fso_unload_event.0.location.country.source',
            'events.cargo_fso_unload_event.0.location.region.id',
            'events.cargo_fso_unload_event.0.location.region.label',
            'events.cargo_fso_unload_event.0.location.region.layer',
            'events.cargo_fso_unload_event.0.location.region.probability',
            'events.cargo_fso_unload_event.0.location.region.source',
            'events.cargo_fso_unload_event.0.location.shipping_region.id',
            'events.cargo_fso_unload_event.0.location.shipping_region.label',
            'events.cargo_fso_unload_event.0.location.shipping_region.layer',
            'events.cargo_fso_unload_event.0.location.shipping_region.probability',
            'events.cargo_fso_unload_event.0.location.shipping_region.source',
            'events.cargo_fso_unload_event.0.location.sts_zone.id',
            'events.cargo_fso_unload_event.0.location.sts_zone.label',
            'events.cargo_fso_unload_event.0.location.sts_zone.layer',
            'events.cargo_fso_unload_event.0.location.sts_zone.probability',
            'events.cargo_fso_unload_event.0.location.sts_zone.source',
            'events.cargo_fso_unload_event.0.location.trading_block.id',
            'events.cargo_fso_unload_event.0.location.trading_block.label',
            'events.cargo_fso_unload_event.0.location.trading_block.layer',
            'events.cargo_fso_unload_event.0.location.trading_block.probability',
            'events.cargo_fso_unload_event.0.location.trading_block.source',
            'events.cargo_fso_unload_event.0.location.trading_region.id',
            'events.cargo_fso_unload_event.0.location.trading_region.label',
            'events.cargo_fso_unload_event.0.location.trading_region.layer',
            'events.cargo_fso_unload_event.0.location.trading_region.probability',
            'events.cargo_fso_unload_event.0.location.trading_region.source',
            'events.cargo_fso_unload_event.0.location.trading_subregion.id',
            'events.cargo_fso_unload_event.0.location.trading_subregion.label',
            'events.cargo_fso_unload_event.0.location.trading_subregion.layer',
            'events.cargo_fso_unload_event.0.location.trading_subregion.probability',
            'events.cargo_fso_unload_event.0.location.trading_subregion.source',
            'events.cargo_fso_unload_event.0.pos.0',
            'events.cargo_fso_unload_event.0.pos.1',
            'events.cargo_fso_unload_event.0.probability',
            'events.cargo_fso_unload_event.0.start_timestamp',
            'events.cargo_port_load_event.0.end_timestamp',
            'events.cargo_port_load_event.0.event_type',
            'events.cargo_port_load_event.0.location.country.id',
            'events.cargo_port_load_event.0.location.country.label',
            'events.cargo_port_load_event.0.location.country.layer',
            'events.cargo_port_load_event.0.location.country.probability',
            'events.cargo_port_load_event.0.location.country.source',
            'events.cargo_port_load_event.0.location.port.id',
            'events.cargo_port_load_event.0.location.port.label',
            'events.cargo_port_load_event.0.location.port.layer',
            'events.cargo_port_load_event.0.location.port.probability',
            'events.cargo_port_load_event.0.location.port.source',
            'events.cargo_port_load_event.0.location.region.id',
            'events.cargo_port_load_event.0.location.region.label',
            'events.cargo_port_load_event.0.location.region.layer',
            'events.cargo_port_load_event.0.location.region.probability',
            'events.cargo_port_load_event.0.location.region.source',
            'events.cargo_port_load_event.0.location.shipping_region.id',
            'events.cargo_port_load_event.0.location.shipping_region.label',
            'events.cargo_port_load_event.0.location.shipping_region.layer',
            'events.cargo_port_load_event.0.location.shipping_region.probability',
            'events.cargo_port_load_event.0.location.shipping_region.source',
            'events.cargo_port_load_event.0.location.terminal.id',
            'events.cargo_port_load_event.0.location.terminal.label',
            'events.cargo_port_load_event.0.location.terminal.layer',
            'events.cargo_port_load_event.0.location.terminal.probability',
            'events.cargo_port_load_event.0.location.terminal.source',
            'events.cargo_port_load_event.0.location.trading_block.id',
            'events.cargo_port_load_event.0.location.trading_block.label',
            'events.cargo_port_load_event.0.location.trading_block.layer',
            'events.cargo_port_load_event.0.location.trading_block.probability',
            'events.cargo_port_load_event.0.location.trading_block.source',
            'events.cargo_port_load_event.0.location.trading_region.id',
            'events.cargo_port_load_event.0.location.trading_region.label',
            'events.cargo_port_load_event.0.location.trading_region.layer',
            'events.cargo_port_load_event.0.location.trading_region.probability',
            'events.cargo_port_load_event.0.location.trading_region.source',
            'events.cargo_port_load_event.0.location.trading_subregion.id',
            'events.cargo_port_load_event.0.location.trading_subregion.label',
            'events.cargo_port_load_event.0.location.trading_subregion.layer',
            'events.cargo_port_load_event.0.location.trading_subregion.probability',
            'events.cargo_port_load_event.0.location.trading_subregion.source',
            'events.cargo_port_load_event.0.pos.0',
            'events.cargo_port_load_event.0.pos.1',
            'events.cargo_port_load_event.0.probability',
            'events.cargo_port_load_event.0.start_timestamp',
            'events.cargo_port_unload_event.0.end_timestamp',
            'events.cargo_port_unload_event.0.event_type',
            'events.cargo_port_unload_event.0.location.country.id',
            'events.cargo_port_unload_event.0.location.country.label',
            'events.cargo_port_unload_event.0.location.country.layer',
            'events.cargo_port_unload_event.0.location.country.probability',
            'events.cargo_port_unload_event.0.location.country.source',
            'events.cargo_port_unload_event.0.location.port.id',
            'events.cargo_port_unload_event.0.location.port.label',
            'events.cargo_port_unload_event.0.location.port.layer',
            'events.cargo_port_unload_event.0.location.port.probability',
            'events.cargo_port_unload_event.0.location.port.source',
            'events.cargo_port_unload_event.0.location.region.id',
            'events.cargo_port_unload_event.0.location.region.label',
            'events.cargo_port_unload_event.0.location.region.layer',
            'events.cargo_port_unload_event.0.location.region.probability',
            'events.cargo_port_unload_event.0.location.region.source',
            'events.cargo_port_unload_event.0.location.shipping_region.id',
            'events.cargo_port_unload_event.0.location.shipping_region.label',
            'events.cargo_port_unload_event.0.location.shipping_region.layer',
            'events.cargo_port_unload_event.0.location.shipping_region.probability',
            'events.cargo_port_unload_event.0.location.shipping_region.source',
            'events.cargo_port_unload_event.0.location.sts_zone.id',
            'events.cargo_port_unload_event.0.location.sts_zone.label',
            'events.cargo_port_unload_event.0.location.sts_zone.layer',
            'events.cargo_port_unload_event.0.location.sts_zone.probability',
            'events.cargo_port_unload_event.0.location.sts_zone.source',
            'events.cargo_port_unload_event.0.location.terminal.id',
            'events.cargo_port_unload_event.0.location.terminal.label',
            'events.cargo_port_unload_event.0.location.terminal.layer',
            'events.cargo_port_unload_event.0.location.terminal.probability',
            'events.cargo_port_unload_event.0.location.terminal.source',
            'events.cargo_port_unload_event.0.location.trading_block.id',
            'events.cargo_port_unload_event.0.location.trading_block.label',
            'events.cargo_port_unload_event.0.location.trading_block.layer',
            'events.cargo_port_unload_event.0.location.trading_block.probability',
            'events.cargo_port_unload_event.0.location.trading_block.source',
            'events.cargo_port_unload_event.0.location.trading_region.id',
            'events.cargo_port_unload_event.0.location.trading_region.label',
            'events.cargo_port_unload_event.0.location.trading_region.layer',
            'events.cargo_port_unload_event.0.location.trading_region.probability',
            'events.cargo_port_unload_event.0.location.trading_region.source',
            'events.cargo_port_unload_event.0.location.trading_subregion.id',
            'events.cargo_port_unload_event.0.location.trading_subregion.label',
            'events.cargo_port_unload_event.0.location.trading_subregion.layer',
            'events.cargo_port_unload_event.0.location.trading_subregion.probability',
            'events.cargo_port_unload_event.0.location.trading_subregion.source',
            'events.cargo_port_unload_event.0.pos.0',
            'events.cargo_port_unload_event.0.pos.1',
            'events.cargo_port_unload_event.0.probability',
            'events.cargo_port_unload_event.0.start_timestamp',
            'events.cargo_storage_event.0.end_timestamp',
            'events.cargo_storage_event.0.event_type',
            'events.cargo_storage_event.0.location.country.id',
            'events.cargo_storage_event.0.location.country.label',
            'events.cargo_storage_event.0.location.country.layer',
            'events.cargo_storage_event.0.location.country.probability',
            'events.cargo_storage_event.0.location.country.source',
            'events.cargo_storage_event.0.location.region.id',
            'events.cargo_storage_event.0.location.region.label',
            'events.cargo_storage_event.0.location.region.layer',
            'events.cargo_storage_event.0.location.region.probability',
            'events.cargo_storage_event.0.location.region.source',
            'events.cargo_storage_event.0.location.shipping_region.id',
            'events.cargo_storage_event.0.location.shipping_region.label',
            'events.cargo_storage_event.0.location.shipping_region.layer',
            'events.cargo_storage_event.0.location.shipping_region.probability',
            'events.cargo_storage_event.0.location.shipping_region.source',
            'events.cargo_storage_event.0.location.trading_block.id',
            'events.cargo_storage_event.0.location.trading_block.label',
            'events.cargo_storage_event.0.location.trading_block.layer',
            'events.cargo_storage_event.0.location.trading_block.probability',
            'events.cargo_storage_event.0.location.trading_block.source',
            'events.cargo_storage_event.0.location.trading_region.id',
            'events.cargo_storage_event.0.location.trading_region.label',
            'events.cargo_storage_event.0.location.trading_region.layer',
            'events.cargo_storage_event.0.location.trading_region.probability',
            'events.cargo_storage_event.0.location.trading_region.source',
            'events.cargo_storage_event.0.location.trading_subregion.id',
            'events.cargo_storage_event.0.location.trading_subregion.label',
            'events.cargo_storage_event.0.location.trading_subregion.layer',
            'events.cargo_storage_event.0.location.trading_subregion.probability',
            'events.cargo_storage_event.0.location.trading_subregion.source',
            'events.cargo_storage_event.0.pos.0',
            'events.cargo_storage_event.0.pos.1',
            'events.cargo_storage_event.0.start_timestamp',
            'events.cargo_storage_event.0.vessel_id',
            'events.cargo_sts_event.0.end_timestamp',
            'events.cargo_sts_event.0.event_type',
            'events.cargo_sts_event.0.from_vessel_id',
            'events.cargo_sts_event.0.from_vessel_name',
            'events.cargo_sts_event.0.location.country.id',
            'events.cargo_sts_event.0.location.country.label',
            'events.cargo_sts_event.0.location.country.layer',
            'events.cargo_sts_event.0.location.country.probability',
            'events.cargo_sts_event.0.location.country.source',
            'events.cargo_sts_event.0.location.port.id',
            'events.cargo_sts_event.0.location.port.label',
            'events.cargo_sts_event.0.location.port.layer',
            'events.cargo_sts_event.0.location.port.probability',
            'events.cargo_sts_event.0.location.port.source',
            'events.cargo_sts_event.0.location.region.id',
            'events.cargo_sts_event.0.location.region.label',
            'events.cargo_sts_event.0.location.region.layer',
            'events.cargo_sts_event.0.location.region.probability',
            'events.cargo_sts_event.0.location.region.source',
            'events.cargo_sts_event.0.location.shipping_region.id',
            'events.cargo_sts_event.0.location.shipping_region.label',
            'events.cargo_sts_event.0.location.shipping_region.layer',
            'events.cargo_sts_event.0.location.shipping_region.probability',
            'events.cargo_sts_event.0.location.shipping_region.source',
            'events.cargo_sts_event.0.location.sts_zone.id',
            'events.cargo_sts_event.0.location.sts_zone.label',
            'events.cargo_sts_event.0.location.sts_zone.layer',
            'events.cargo_sts_event.0.location.sts_zone.probability',
            'events.cargo_sts_event.0.location.sts_zone.source',
            'events.cargo_sts_event.0.location.trading_block.id',
            'events.cargo_sts_event.0.location.trading_block.label',
            'events.cargo_sts_event.0.location.trading_block.layer',
            'events.cargo_sts_event.0.location.trading_block.probability',
            'events.cargo_sts_event.0.location.trading_block.source',
            'events.cargo_sts_event.0.location.trading_region.id',
            'events.cargo_sts_event.0.location.trading_region.label',
            'events.cargo_sts_event.0.location.trading_region.layer',
            'events.cargo_sts_event.0.location.trading_region.probability',
            'events.cargo_sts_event.0.location.trading_region.source',
            'events.cargo_sts_event.0.location.trading_subregion.id',
            'events.cargo_sts_event.0.location.trading_subregion.label',
            'events.cargo_sts_event.0.location.trading_subregion.layer',
            'events.cargo_sts_event.0.location.trading_subregion.probability',
            'events.cargo_sts_event.0.location.trading_subregion.source',
            'events.cargo_sts_event.0.pos.0',
            'events.cargo_sts_event.0.pos.1',
            'events.cargo_sts_event.0.start_timestamp',
            'events.cargo_sts_event.0.to_vessel_id',
            'events.cargo_sts_event.0.to_vessel_name',
            'events.cargo_sts_event.1.end_timestamp',
            'events.cargo_sts_event.1.event_type',
            'events.cargo_sts_event.1.from_vessel_id',
            'events.cargo_sts_event.1.from_vessel_name',
            'events.cargo_sts_event.1.location.country.id',
            'events.cargo_sts_event.1.location.country.label',
            'events.cargo_sts_event.1.location.country.layer',
            'events.cargo_sts_event.1.location.country.probability',
            'events.cargo_sts_event.1.location.country.source',
            'events.cargo_sts_event.1.location.region.id',
            'events.cargo_sts_event.1.location.region.label',
            'events.cargo_sts_event.1.location.region.layer',
            'events.cargo_sts_event.1.location.region.probability',
            'events.cargo_sts_event.1.location.region.source',
            'events.cargo_sts_event.1.location.shipping_region.id',
            'events.cargo_sts_event.1.location.shipping_region.label',
            'events.cargo_sts_event.1.location.shipping_region.layer',
            'events.cargo_sts_event.1.location.shipping_region.probability',
            'events.cargo_sts_event.1.location.shipping_region.source',
            'events.cargo_sts_event.1.location.sts_zone.id',
            'events.cargo_sts_event.1.location.sts_zone.label',
            'events.cargo_sts_event.1.location.sts_zone.layer',
            'events.cargo_sts_event.1.location.sts_zone.probability',
            'events.cargo_sts_event.1.location.sts_zone.source',
            'events.cargo_sts_event.1.location.trading_block.id',
            'events.cargo_sts_event.1.location.trading_block.label',
            'events.cargo_sts_event.1.location.trading_block.layer',
            'events.cargo_sts_event.1.location.trading_block.probability',
            'events.cargo_sts_event.1.location.trading_block.source',
            'events.cargo_sts_event.1.location.trading_region.id',
            'events.cargo_sts_event.1.location.trading_region.label',
            'events.cargo_sts_event.1.location.trading_region.layer',
            'events.cargo_sts_event.1.location.trading_region.probability',
            'events.cargo_sts_event.1.location.trading_region.source',
            'events.cargo_sts_event.1.location.trading_subregion.id',
            'events.cargo_sts_event.1.location.trading_subregion.label',
            'events.cargo_sts_event.1.location.trading_subregion.layer',
            'events.cargo_sts_event.1.location.trading_subregion.probability',
            'events.cargo_sts_event.1.location.trading_subregion.source',
            'events.cargo_sts_event.1.pos.0',
            'events.cargo_sts_event.1.pos.1',
            'events.cargo_sts_event.1.start_timestamp',
            'events.cargo_sts_event.1.to_vessel_id',
            'events.cargo_sts_event.1.to_vessel_name',
            'product.category.id',
            'product.category.label',
            'product.category.layer',
            'product.category.probability',
            'product.category.source',
            'product.grade.id',
            'product.grade.label',
            'product.grade.layer',
            'product.grade.probability',
            'product.grade.source',
            'product.group.id',
            'product.group.label',
            'product.group.layer',
            'product.group.probability',
            'product.group.source',
            'product.group_product.id',
            'product.group_product.label',
            'product.group_product.layer',
            'product.group_product.probability',
            'product.group_product.source',
            'quantity',
            'status',
            'vessels.0.corporate_entities.charterer.id',
            'vessels.0.corporate_entities.charterer.label',
            'vessels.0.corporate_entities.charterer.layer',
            'vessels.0.corporate_entities.charterer.probability',
            'vessels.0.corporate_entities.charterer.source',
            'vessels.0.corporate_entities.commercial_owner.id',
            'vessels.0.corporate_entities.commercial_owner.label',
            'vessels.0.corporate_entities.commercial_owner.layer',
            'vessels.0.corporate_entities.commercial_owner.probability',
            'vessels.0.corporate_entities.commercial_owner.source',
            'vessels.0.corporate_entities.time_charterer.end_timestamp',
            'vessels.0.corporate_entities.time_charterer.id',
            'vessels.0.corporate_entities.time_charterer.label',
            'vessels.0.corporate_entities.time_charterer.layer',
            'vessels.0.corporate_entities.time_charterer.probability',
            'vessels.0.corporate_entities.time_charterer.source',
            'vessels.0.corporate_entities.time_charterer.start_timestamp',
            'vessels.0.cubic_capacity',
            'vessels.0.dwt',
            'vessels.0.end_timestamp',
            'vessels.0.fixture_fulfilled',
            'vessels.0.fixture_id',
            'vessels.0.id',
            'vessels.0.imo',
            'vessels.0.mmsi',
            'vessels.0.name',
            'vessels.0.start_timestamp',
            'vessels.0.status',
            'vessels.0.tags.0.end_timestamp',
            'vessels.0.tags.0.start_timestamp',
            'vessels.0.tags.0.tag',
            'vessels.0.vessel_class',
            'vessels.0.voyage_id',
            'vessels.1.corporate_entities.charterer.id',
            'vessels.1.corporate_entities.charterer.label',
            'vessels.1.corporate_entities.charterer.layer',
            'vessels.1.corporate_entities.charterer.probability',
            'vessels.1.corporate_entities.charterer.source',
            'vessels.1.corporate_entities.commercial_owner.id',
            'vessels.1.corporate_entities.commercial_owner.label',
            'vessels.1.corporate_entities.commercial_owner.layer',
            'vessels.1.corporate_entities.commercial_owner.probability',
            'vessels.1.corporate_entities.commercial_owner.source',
            'vessels.1.corporate_entities.time_charterer.end_timestamp',
            'vessels.1.corporate_entities.time_charterer.id',
            'vessels.1.corporate_entities.time_charterer.label',
            'vessels.1.corporate_entities.time_charterer.layer',
            'vessels.1.corporate_entities.time_charterer.probability',
            'vessels.1.corporate_entities.time_charterer.source',
            'vessels.1.corporate_entities.time_charterer.start_timestamp',
            'vessels.1.cubic_capacity',
            'vessels.1.dwt',
            'vessels.1.end_timestamp',
            'vessels.1.fixture_fulfilled',
            'vessels.1.id',
            'vessels.1.imo',
            'vessels.1.mmsi',
            'vessels.1.name',
            'vessels.1.start_timestamp',
            'vessels.1.status',
            'vessels.1.tags.0.end_timestamp',
            'vessels.1.tags.0.start_timestamp',
            'vessels.1.tags.0.tag',
            'vessels.1.vessel_class',
            'vessels.1.voyage_id',
            'vessels.2.corporate_entities.charterer.id',
            'vessels.2.corporate_entities.charterer.label',
            'vessels.2.corporate_entities.charterer.layer',
            'vessels.2.corporate_entities.charterer.probability',
            'vessels.2.corporate_entities.charterer.source',
            'vessels.2.corporate_entities.commercial_owner.id',
            'vessels.2.corporate_entities.commercial_owner.label',
            'vessels.2.corporate_entities.commercial_owner.layer',
            'vessels.2.corporate_entities.commercial_owner.probability',
            'vessels.2.corporate_entities.commercial_owner.source',
            'vessels.2.corporate_entities.time_charterer.end_timestamp',
            'vessels.2.corporate_entities.time_charterer.id',
            'vessels.2.corporate_entities.time_charterer.label',
            'vessels.2.corporate_entities.time_charterer.layer',
            'vessels.2.corporate_entities.time_charterer.probability',
            'vessels.2.corporate_entities.time_charterer.source',
            'vessels.2.corporate_entities.time_charterer.start_timestamp',
            'vessels.2.cubic_capacity',
            'vessels.2.dwt',
            'vessels.2.end_timestamp',
            'vessels.2.id',
            'vessels.2.imo',
            'vessels.2.mmsi',
            'vessels.2.name',
            'vessels.2.start_timestamp',
            'vessels.2.status',
            'vessels.2.tags.0.start_timestamp',
            'vessels.2.tags.0.tag',
            'vessels.2.vessel_class',
            'vessels.2.voyage_id',
            'vessels.3.corporate_entities.commercial_owner.id',
            'vessels.3.corporate_entities.commercial_owner.label',
            'vessels.3.corporate_entities.commercial_owner.layer',
            'vessels.3.corporate_entities.commercial_owner.probability',
            'vessels.3.corporate_entities.commercial_owner.source',
            'vessels.3.cubic_capacity',
            'vessels.3.dwt',
            'vessels.3.id',
            'vessels.3.imo',
            'vessels.3.mmsi',
            'vessels.3.name',
            'vessels.3.start_timestamp',
            'vessels.3.status',
            'vessels.3.vessel_class',
            'vessels.3.voyage_id',
            'parent_ids.0.id',
            'parent_ids.0.splinter_timestamp',
            'parent_ids.1.id',
            'parent_ids.1.splinter_timestamp',
        ]
        ```

        """
        if columns is None:
            columns = DEFAULT_COLUMNS

        flatten = functools.partial(
            convert_cargo_movement_to_flat_dict, cols=columns
        )

        logger.debug("Converting each CargoMovement to a flat dictionary")
        with Pool(os.cpu_count()) as pool:
            records = pool.map(flatten, super().to_list())

        return create_dataframe(
            columns=columns,
            default_columns=DEFAULT_COLUMNS,
            data=records,
            logger_description="CargoMovements",
        )


DEFAULT_COLUMNS = [
    "events.cargo_port_load_event.0.location.port.label",
    "events.cargo_port_unload_event.0.location.port.label",
    "product.group.label",
    "product.grade.label",
    "quantity",
    "vessels.0.name",
    "events.cargo_port_load_event.0.end_timestamp",
    "events.cargo_port_unload_event.0.start_timestamp",
]



"""Vessel Movements Endpoint."""
from datetime import datetime
from typing import List, Union

from vortexasdk.api import ID
from vortexasdk.api.shared_types import to_ISODate
from vortexasdk.endpoints.endpoints import VESSEL_MOVEMENTS_RESOURCE
from vortexasdk.endpoints.vessel_movements_result import VesselMovementsResult
from vortexasdk.logger import get_logger
from vortexasdk.operations import Search
from vortexasdk.utils import convert_to_list

logger = get_logger(__name__)


class VesselMovements(Search):
    """
    Vessel Movements Endpoint, use this to search through Vortexa's VesselMovements.

    A VesselMovement represents a single vessel moving between two locations.

    * The vessel may carry one cargo, many cargoes (coloads), or zero cargos (ballast).
    * The start and end locations for a VesselMovement may be on land (loadings and discharges), they may be STS Zones
    (STS events), or they may be Floating Storage.

    A detailed explanation of Cargo/Vessel Movements can be found [here](https://docs.vortexa.com/reference/intro-movement-difference).
    """

    _MAX_PAGE_RESULT_SIZE = 500

    def __init__(self):
        Search.__init__(self, VESSEL_MOVEMENTS_RESOURCE)

    def search(
        self,
        filter_time_min: datetime = datetime(2019, 10, 1, 0),
        filter_time_max: datetime = datetime(2019, 10, 1, 1),
        unit: str = "b",
        filter_activity: str = None,
        filter_charterers: Union[ID, List[ID]] = None,
        filter_destinations: Union[ID, List[ID]] = None,
        filter_origins: Union[ID, List[ID]] = None,
        filter_owners: Union[ID, List[ID]] = None,
        filter_products: Union[ID, List[ID]] = None,
        filter_vessels: Union[ID, List[ID]] = None,
        filter_vessel_classes: Union[ID, List[ID]] = None,
        filter_vessel_status: str = None,
        filter_vessel_age_min: int = None,
        filter_vessel_age_max: int = None,
        filter_vessel_scrubbers: str = "disabled",
        filter_vessel_flags: Union[ID, List[ID]] = None,
        filter_vessel_ice_class: Union[ID, List[ID]] = None,
        filter_vessel_propulsion: Union[ID, List[ID]] = None,
        exclude_origins: Union[ID, List[ID]] = None,
        exclude_destinations: Union[ID, List[ID]] = None,
        exclude_products: Union[ID, List[ID]] = None,
        exclude_vessels: Union[ID, List[ID]] = None,
        exclude_vessel_classes: Union[ID, List[ID]] = None,
        exclude_charterers: Union[ID, List[ID]] = None,
        exclude_owners: Union[ID, List[ID]] = None,
        exclude_vessel_flags: Union[ID, List[ID]] = None,
        exclude_vessel_ice_class: Union[ID, List[ID]] = None,
        exclude_vessel_propulsion: Union[ID, List[ID]] = None,
    ) -> VesselMovementsResult:
        """
        Find VesselMovements matching the given search parameters.

        # Arguments
            filter_activity: Movement activity on which to base the time filter. Must be one of ['loading_state',
             'loading_start', 'loading_end', 'identified_for_loading_state', 'unloading_state', 'unloading_start',
              'unloading_end', 'unloaded_state', 'storing_state', 'storing_start', 'storing_end', 'transiting_state',
               'any_activity'].

            filter_time_min: The UTC start date of the time filter.

            filter_time_max: The UTC end date of the time filter.

            unit: Unit of measurement. Enter 'b' for barrels or 't' for tonnes.

            filter_charterers: A charterer ID, or list of charterer IDs to filter on.

            filter_destinations: A geography ID, or list of geography IDs to filter on.

            filter_origins: A geography ID, or list of geography IDs to filter on.

            filter_owners: An corporation ID, or list of corporation IDs to filter on.

            filter_products: A product ID, or list of product IDs to filter on.

            filter_vessels: A vessel ID, or list of vessel IDs to filter on.

            filter_vessel_classes: A vessel class, or list of vessel classes to filter on.

            filter_vessel_status: The vessel status on which to base the filter. Enter 'vessel_status_ballast' for ballast vessels, 'vessel_status_laden_known' for laden vessels with known cargo (i.e. a type of cargo that Vortexa currently tracks) or 'vessel_status_laden_unknown' for laden vessels with unknown cargo (i.e. a type of cargo that Vortexa currently does not track).

            filter_vessel_age_min: A number between 1 and 100 (representing years).

            filter_vessel_age_max: A number between 1 and 100 (representing years).

            filter_vessel_scrubbers: Either inactive 'disabled', or included 'inc' or excluded 'exc'.

            filter_vessel_flags: A geography ID, or list of geography IDs to filter on.

            filter_vessel_ice_class: An attribute ID, or list of attribute IDs to filter on.

            filter_vessel_propulsion: An attribute ID, or list of attribute IDs to filter on.

            exclude_origins: A geography ID, or list of geography IDs to exclude.

            exclude_destinations: A geography ID, or list of geography IDs to exclude.

            exclude_products: A product ID, or list of product IDs to exclude.

            exclude_vessels: A vessel ID, or list of vessel IDs to exclude.

            exclude_vessel_classes: A vessel class, or list of vessel classes to exclude.

            exclude_charterers: A charterer ID, or list of charterer IDs to exclude.

            exclude_owners: An owner ID, or list of owner IDs to exclude.

            exclude_vessel_flags: A geography ID, or list of geography IDs to exclude.

            exclude_vessel_ice_class: An attribute ID, or list of attribute IDs to exclude.

            exclude_vessel_propulsion: An attribute ID, or list of attribute IDs to exclude.


        # Returns
        `VesselMovementsResult`, containing all the vessel movements matching the given search terms.


        # Example
        Let's search for all vessels that departed from `Rotterdam [NL]` on the morning of 1st December 2018.

        ```python
        >>> from vortexasdk import VesselMovements, Geographies
        >>> rotterdam = [g.id for g in Geographies().search("rotterdam").to_list() if "port" in g.layer]
        >>> df = VesselMovements().search(
        ...        filter_time_min=datetime(2017, 10, 1, 0, 0),
        ...        filter_time_max=datetime(2017, 10, 1, 0, 10),
        ...        filter_origins=rotterdam
        ... ).to_df().head(2)

        ```

        |    | start_timestamp          | end_timestamp            |   vessel.imo | vessel.name   | vessel.vessel_class   | origin.location.country.label   | origin.location.port.label   | destination.location.country.label   | destination.location.port.label   |   cargoes.0.quantity | cargoes.0.product.grade.label   |
        |---:|:-------------------------|:-------------------------|-------------:|:--------------|:----------------------|:--------------------------------|:-----------------------------|:-------------------------------------|:----------------------------------|---------------------:|:--------------------------------|
        |  0 | 2017-09-30T15:30:27+0000 | 2017-10-03T01:46:06+0000 |  9.21091e+06 | ADEBOMI 3     | handysize             | Netherlands                     | Rotterdam [NL]               | Netherlands                          | Rotterdam [NL]                    |                  nan | nan                             |
        |  1 | 2017-08-29T14:51:32+0000 | 2017-10-04T14:46:21+0000 |  9.64544e+06 | AEGEAN VISION | suezmax               | Netherlands                     | Rotterdam [NL]               | Singapore                            | Singapore [SG]                    |               852261 | High Sulphur                    |

        [Vessel Movements Endpoint Further Documentation](https://docs.vortexa.com/reference/POST/vessel-movements/search)

        """
        exclude_params = {
            "filter_origins": convert_to_list(exclude_origins),
            "filter_destinations": convert_to_list(exclude_destinations),
            "filter_products": convert_to_list(exclude_products),
            "filter_vessels": convert_to_list(exclude_vessels),
            "filter_vessel_classes": convert_to_list(exclude_vessel_classes),
            "filter_charterers": convert_to_list(exclude_charterers),
            "filter_owners": convert_to_list(exclude_owners),
            "filter_vessel_flags": convert_to_list(exclude_vessel_flags),
            "filter_vessel_ice_class": convert_to_list(
                exclude_vessel_ice_class
            ),
            "filter_vessel_propulsion": convert_to_list(
                exclude_vessel_propulsion
            ),
        }

        api_params = {
            "filter_activity": filter_activity,
            "filter_time_min": to_ISODate(filter_time_min),
            "filter_time_max": to_ISODate(filter_time_max),
            "unit": unit,
            "filter_charterers": convert_to_list(filter_charterers),
            "filter_owners": convert_to_list(filter_owners),
            "filter_destinations": convert_to_list(filter_destinations),
            "filter_origins": convert_to_list(filter_origins),
            "filter_products": convert_to_list(filter_products),
            "filter_vessels": convert_to_list(filter_vessels),
            "filter_vessel_classes": convert_to_list(filter_vessel_classes),
            "filter_vessel_status": filter_vessel_status,
            "filter_vessel_age_min": filter_vessel_age_min,
            "filter_vessel_age_max": filter_vessel_age_max,
            "filter_vessel_scrubbers": filter_vessel_scrubbers,
            "filter_vessel_flags": convert_to_list(filter_vessel_flags),
            "filter_vessel_ice_class": convert_to_list(
                filter_vessel_ice_class
            ),
            "filter_vessel_propulsion": convert_to_list(
                filter_vessel_propulsion
            ),
            "exclude": exclude_params,
            "size": self._MAX_PAGE_RESULT_SIZE,
        }

        return VesselMovementsResult(super().search(**api_params))



from unittest import TestCase
from distutils.version import LooseVersion

from vortexasdk.version_utils import (
    get_latest_sdk_version,
)


class TestVersionUtils(TestCase):
    def test_get_latest_sdk_version(self):
        sdk_old_version = "0.18.0"

        lv = get_latest_sdk_version()

        assert lv > LooseVersion(sdk_old_version)



from dataclasses import dataclass
from unittest import TestCase

from vortexasdk.api.serdes import FromDictMixin


@dataclass
class Foo(FromDictMixin):
    bar: int


class TestFromDictMixin(TestCase):
    def test_from_dict(self):
        d = {"bar": 1}

        actual = Foo.from_dict(d)

        expected = Foo(1)

        assert actual == expected



from itertools import groupby
from typing import Dict, List

# noinspection PyProtectedMember
from flatten_dict import flatten


def _group_by_layer(entity_list: List[Dict]) -> Dict:
    return {e["layer"]: e for e in entity_list}


def _format_keys(dictionary):
    flat_with_formatted_keys = {}
    for k, v in dictionary.items():
        nice_path = ".".join([str(i) for i in k])
        flat_with_formatted_keys[nice_path] = v
    return flat_with_formatted_keys


def _serialize_ve_layer(ve: Dict) -> Dict:
    ve["corporate_entities"] = _group_by_layer(ve["corporate_entities"])
    return ve


def _serialize_ce_layer(ce: Dict) -> Dict:
    ce["location"] = _group_by_layer(ce["location"])
    return ce


def _group_cme_attributes_by_layer(cme: Dict) -> Dict:
    """Group relevant CargoMovementEntity attributes by `Entity.layer`."""
    vessels = [_serialize_ve_layer(ve) for ve in cme["vessels"]]
    products = _group_by_layer(cme["product"])

    events = {
        event_type: list(g)
        for event_type, g in groupby(cme["events"], lambda x: x["event_type"])
    }

    events_attributes = {
        event_type: [_serialize_ce_layer(ce) for ce in es]
        for event_type, es in events.items()
    }

    cme["product"] = products
    cme["vessels"] = vessels
    cme["events"] = events_attributes
    return cme


def convert_cme_to_flat_dict(cme: Dict, cols="all") -> Dict:
    """Convert nested `CargoMovementEntity` object to flat dictionary, keeping *cols*."""
    as_dict = _group_cme_attributes_by_layer(cme)

    formatted = flatten_dictionary(as_dict)

    if cols == "all":
        return formatted
    else:
        return {k: v for k, v in formatted.items() if k in cols}


def flatten_dictionary(d: Dict) -> Dict:
    """Flatten dictionary, then format keys."""
    return _format_keys(flatten(d, enumerate_types=(list,)))



"""Products Endpoint."""
from typing import Dict, List, Union

from vortexasdk.api.shared_types import ID
from vortexasdk.endpoints.endpoints import PRODUCTS_REFERENCE
from vortexasdk.endpoints.products_result import ProductResult
from vortexasdk.operations import Reference, Search
from vortexasdk.utils import convert_to_list


class Products(Reference, Search):
    """Products endpoint."""

    def __init__(self):
        """Instantiate endpoint using reference endpoint."""
        Reference.__init__(self, PRODUCTS_REFERENCE)
        Search.__init__(self, PRODUCTS_REFERENCE)

    def load_all(self) -> ProductResult:
        """Load all products."""
        return self.search()

    def search(
        self,
        term: Union[str, List[str]] = None,
        ids: Union[str, List[str]] = None,
        product_parent: Union[str, List[str]] = None,
        exact_term_match: bool = False,
    ) -> ProductResult:
        """
        Find all products matching given search terms.

        # Arguments
            term: The name(s) (or partial name(s)) of a product we'd like to search

            ids: ID or IDs of products we'd like to search

            product_parent: ID, or list of IDs of the immediate product parent. E.g. `product_parent ='12345'` will return all children of product `12345`.

            exact_term_match: By default, the SDK returns all products which name _includes_ the search term. For example, searching for "Gasoil" will return
                results including "Gasoil", "Gasoil 0.4pc", "Gasoil 500ppm" etc. Setting `exact_search_match` to true ensure that only exact term matches are
                returned, ie just "Gasoil" in this case.

        # Returns
        List of products matching the search arguments.


        # Examples

        Let's look for products with in one of `['diesel', 'fuel oil', 'grane']` their name, or related names.

        ```python
        >>> from vortexasdk import Products
        >>> df = Products().search(term=['diesel', 'fuel oil', 'grane']).to_df('all')

        ```
        Returns

        |    | id                 | name          | layer.0   | leaf   | parent.0.name   | parent.0.layer.0   | parent.0.id       |   meta.api_min |   meta.api_max | ref_type   |   meta.sulphur_min |   meta.sulphur_max |
        |---:|:-------------------|:--------------|:----------|:-------|:----------------|:-------------------|:------------------|---------------:|---------------:|:-----------|-------------------:|-------------------:|
        |  0 | 1c107b4317bc2c8... | Fuel Oil      | category  | False  | Dirty products  | product            | 5de0b00094e0fd... |        12.8878 |        12.8878 | product    |             nan    |             nan    |
        |  1 | fddedd17e02507f... | Grane         | grade     | True   | Medium-Sour     | subproduct_group   | a7e26956fbb917... |        29.2955 |        29.2955 | product    |               0.62 |               0.62 |
        |  2 | deda35eb9ca56b5... | Diesel/Gasoil | category  | False  | Clean products  | product            | b68cbb7746f8b9... |        35.9556 |        35.9556 | product    |             nan    |             nan    |


        # Further Documentation

        [VortexaAPI Product Reference](https://docs.vortexa.com/reference/POST/reference/products)

        """
        api_params = {
            "term": convert_to_list(term),
            "ids": convert_to_list(ids),
            "product_parent": convert_to_list(product_parent),
            "allowTopLevelProducts": True,
        }

        return ProductResult(
            super().search(exact_term_match=exact_term_match, **api_params)
        )

    def reference(self, id: ID) -> Dict:
        """
        Perform a product lookup.

        # Arguments
            id: Product ID to lookup

        # Returns
        Product record matching the ID

        # Further Documentation:
        [VortexaAPI Product Reference](https://docs.vortexa.com/reference/POST/reference/products)

        """
        return super().reference(id)



from typing import List

import pandas as pd

from vortexasdk.api import Product
from vortexasdk.api.entity_flattening import flatten_dictionary
from vortexasdk.api.search_result import Result
from vortexasdk.logger import get_logger
from vortexasdk.result_conversions import create_dataframe, create_list

logger = get_logger(__name__)


class ProductResult(Result):
    """Container class that holds the result obtained from calling the `Product` endpoint."""

    def to_list(self) -> List[Product]:
        """Represent products as a list."""
        # noinspection PyTypeChecker
        return create_list(super().to_list(), Product)

    def to_df(self, columns=None) -> pd.DataFrame:
        """
        Represent products as a `pd.DataFrame`.

        # Arguments
            columns: The product features we want in the dataframe. Enter `columns='all'` to include all features.
            Defaults to `columns = ['id', 'name', 'layer.0', 'parent.0.name']`.


        # Returns
        `pd.DataFrame` of products.

        """
        flattened_dicts = [flatten_dictionary(p) for p in super().to_list()]

        return create_dataframe(
            columns=columns,
            default_columns=DEFAULT_COLUMNS,
            data=flattened_dicts,
            logger_description="Products",
        )


DEFAULT_COLUMNS = ["id", "name", "layer.0", "parent.0.name"]



import logging
import sys
from logging.handlers import TimedRotatingFileHandler

from vortexasdk.config import LOG_FILE, LOG_LEVEL

FORMATTER = logging.Formatter(
    "%(asctime)s %(name)s — %(levelname)s — %(message)s"
)


def get_logger(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(LOG_LEVEL)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    logger.addHandler(console_handler)

    if LOG_FILE is not None:
        file_handler = TimedRotatingFileHandler(LOG_FILE, when="midnight")
        file_handler.setFormatter(FORMATTER)
        logger.addHandler(file_handler)

    logger.propagate = False
    return logger



import json
from urllib.request import urlopen
from distutils.version import LooseVersion
from vortexasdk import __name__ as sdk_pkg_name
from vortexasdk.version import __version__


def get_latest_sdk_version() -> str:
    """Retrieves the latest SDK version from PyPI."""
    url = f"https://pypi.python.org/pypi/{sdk_pkg_name}/json"
    with urlopen(url) as u:
        data = json.loads(u.read())

    versions = data["releases"].keys()
    sorted_versions = sorted(versions, key=LooseVersion)
    latest_version = sorted_versions[-1]

    return latest_version


def is_sdk_version_outdated():
    """Checks whether SDK version is outdated."""
    latest_version = get_latest_sdk_version()
    if LooseVersion(__version__) < latest_version:
        return latest_version, True
    else:
        return latest_version, False





"""Attributes Endpoint."""
from typing import List, Union
from vortexasdk.endpoints.endpoints import ATTRIBUTES_REFERENCE
from vortexasdk.endpoints.attributes_result import AttributeResult
from vortexasdk.operations import Reference, Search
from vortexasdk.utils import convert_to_list


class Attributes(Reference, Search):
    """
        Attributes endpoint.

        An Attribute is a reference value that corresponds to an ID associated with other entities.

        For example, a vessel object from the Vessel reference endpoint may have the following keys:

        ```json
        {
            "ice_class": "b09ed4e2bd6904dd",
            "propulsion": "3ace0e050724707b"
        }
        ```

        These IDs represent attributes which can be found via the Attributes reference endpoint.

        When the attributes endpoint is searched with those ids as parameters:

        ```python
            >>> from vortexasdk import Attributes
            >>> df = Attributes().search(ids=["b09ed4e2bd6904dd", "3ace0e050724707b"]).to_df()

        ```

        Returns

        |    | id               | type       | label    |
        |---:|:-----------------|:-----------|:---------|
        |  0 | b09ed4e2bd6904dd | ice_class  | UNKNOWN  |
        |  1 | 3ace0e050724707b | propulsion | DFDE     |


    """

    def __init__(self):
        Reference.__init__(self, ATTRIBUTES_REFERENCE)
        Search.__init__(self, ATTRIBUTES_REFERENCE)

    def load_all(self) -> AttributeResult:
        """
            Load all attributes.
        """
        return self.search()

    # noinspection PyShadowingBuiltins
    def search(
        self,
        type: str = None,
        term: Union[str, List[str]] = None,
        ids: Union[str, List[str]] = None,
    ) -> AttributeResult:
        """
        Find all attributes matching given type.

        # Arguments
            type: The type of attribute we're filtering on. Type can be: `ice_class`, `propulsion`, `scrubber`

        # Returns
        List of attributes matching `type`


        # Examples

        Find all attributes with a type of `ice_class`.
        ```python
        >>> from vortexasdk import Attributes
        >>> df = Attributes().search(type="scrubber").to_df()

        ```
        returns

        |    | id               | name       | type        |
        |---:|:-----------------|:-----------|:------------|
        |  0 | 14c7b073809eb565 | Open Loop  | scrubber    |
        |  1 | 478fca39000c49d6 | Unkown     | scrubber    |

        """

        search_params = {
            "term": [str(e) for e in convert_to_list(term)],
            "ids": convert_to_list(ids),
            "type": type,
        }

        return AttributeResult(super().search(**search_params))



from datetime import datetime

from tests.testcases import TestCaseUsingRealAPI
from vortexasdk import VesselMovements, Geographies, Corporations, Attributes
from vortexasdk.endpoints import vessel_movements_result


class TestVesselMovementsReal(TestCaseUsingRealAPI):
    def test_search(self):
        rotterdam = [
            g.id
            for g in Geographies().search("rotterdam").to_list()
            if "port" in g.layer
        ]
        v = VesselMovements().search(
            filter_time_min=datetime(2017, 10, 1, 0, 0),
            filter_time_max=datetime(2017, 10, 1, 0, 10),
            filter_origins=rotterdam,
        )

        assert len(v) > 10

    def test_search_vessel_status(self):
        v = VesselMovements().search(
            filter_time_min=datetime(2019, 10, 1),
            filter_time_max=datetime(2019, 10, 15),
            filter_vessel_classes=["vlcc"],
            filter_vessel_status="vessel_status_ballast",
        )

        assert len(v) > 50

    def test_exclusion_filter(self):
        meg = [
            g.id
            for g in Geographies().search("MEG/AG").to_list()
            if "trading_region" in g.layer
        ]
        iraq = [
            g.id
            for g in Geographies().search("Iraq").to_list()
            if "country" in g.layer
        ]
        bahri = [c.id for c in Corporations().search("BAHRI").to_list()]

        cols = [
            "vessel_movement_id",
            "vessel.name",
            "start_timestamp",
            "end_timestamp",
            "origin.location.country.id",
            "origin.location.country.label",
            "destination.location.country.id",
            "destination.location.country.label",
            "cargoes.0.product.group.label",
            "vessel.corporate_entities.charterer.id",
            "vessel.corporate_entities.charterer.label",
        ]

        df = (
            VesselMovements()
            .search(
                filter_origins=meg,
                exclude_origins=iraq,
                exclude_charterers=bahri[0],
                filter_time_min=datetime(2019, 10, 15),
                filter_time_max=datetime(2019, 11, 1),
            )
            .to_df(columns=cols)
        )

        mask = (df["origin.location.country.id"] == iraq[0]) | (
            df["vessel.corporate_entities.charterer.id"] == bahri[0]
        )
        df_excl = df.loc[mask]

        assert df_excl.empty

    def test_to_df_all_columns(self):
        rotterdam = [
            g.id
            for g in Geographies().search("rotterdam").to_list()
            if "port" in g.layer
        ]
        df = (
            VesselMovements()
            .search(
                filter_time_min=datetime(2017, 10, 1, 0, 0),
                filter_time_max=datetime(2017, 10, 1, 0, 10),
                filter_origins=rotterdam,
            )
            .to_df(columns="all")
            .head(2)
        )

        assert len(df) == 2

    def test_search_to_dataframe(self):
        rotterdam = [
            g.id
            for g in Geographies().search("rotterdam").to_list()
            if "port" in g.layer
        ]
        df = (
            VesselMovements()
            .search(
                filter_time_min=datetime(2017, 10, 1, 0, 0),
                filter_time_max=datetime(2017, 10, 1, 0, 10),
                filter_origins=rotterdam,
            )
            .to_df()
            .head(2)
        )

        assert list(df.columns) == vessel_movements_result.DEFAULT_COLUMNS
        assert len(df) == 2

    def test_search_to_dataframe_subset_of_columns(self):
        cols = ["vessel.imo", "vessel.name"]

        rotterdam = [
            g.id
            for g in Geographies().search("rotterdam").to_list()
            if "port" in g.layer
        ]
        df = (
            VesselMovements()
            .search(
                filter_time_min=datetime(2017, 10, 1, 0, 0),
                filter_time_max=datetime(2017, 10, 1, 0, 10),
                filter_origins=rotterdam,
            )
            .to_df(columns=cols)
            .head(2)
        )

        assert list(df.columns) == cols
        assert len(df) == 2

    def test_filter_activity(self):
        df = (
            VesselMovements()
            .search(
                filter_activity="storing_state",
                filter_time_min=datetime(2017, 10, 1),
                filter_time_max=datetime(2017, 10, 1),
            )
            .to_df()
            .head(2)
        )

        print(df.head())
        assert len(df) == 2

    def test_age_flag_scrubbers_filters(self):
        panama = [
            g.id
            for g in Geographies().search("panama").to_list()
            if "country" in g.layer
        ]

        df = (
            VesselMovements()
            .search(
                filter_vessel_scrubbers="inc",
                filter_vessel_age_min=2,
                filter_vessel_age_max=15,
                filter_vessel_flags=panama,
                filter_time_min=datetime(2017, 10, 1),
                filter_time_max=datetime(2017, 10, 1),
            )
            .to_df()
            .head(2)
        )

        print(df.head())
        assert len(df) == 2

    def test_ice_class_filters(self):
        ice_classes = [
            g.id for g in Attributes().search(type="ice_class").to_list()
        ]

        propulsion = [
            g.id for g in Attributes().search(type="propulsion").to_list()
        ]

        df = (
            VesselMovements()
            .search(
                filter_vessel_ice_class=ice_classes,
                filter_vessel_propulsion=propulsion,
                filter_time_min=datetime(2017, 10, 1),
                filter_time_max=datetime(2017, 10, 1),
            )
            .to_df()
            .head(2)
        )

        print(df.head())
        assert len(df) == 2



from typing import Dict, List, Union


def convert_to_list(a) -> List:
    """Convert wraps element in list if element isn't a list already."""
    if a is None:
        return []
    elif isinstance(a, list):
        return a
    else:
        return [a]


def convert_values_to_list(data: Dict) -> Dict:
    """Convert each value to a list."""
    return {k: convert_to_list(v) for k, v in data.items()}


def filter_exact_match(
    allowed_name: Union[str, List[str]], search_result: List[Dict]
) -> List[Dict]:
    """
    Filter search results on items with exact matching names.

    # Arguments
        allowed_name: An allowed name, or list of allowed names
        search_result: A list of dictionaries. Each dictionary must contain the key "name".

    """
    allowed_names_list = convert_to_list(allowed_name)
    return [s for s in search_result if s["name"] in allowed_names_list]


def filter_empty_values(data: Dict) -> Dict:
    return {
        k: v for k, v in data.items() if not (v is None or v == [] or v == {})
    }



"""EIA Forecasts Endpoint."""

from vortexasdk.endpoints.eia_forecasts_result import EIAForecastResult
from vortexasdk.endpoints.endpoints import EIA_FORECASTS_RESOURCE
from vortexasdk.operations import Search
from vortexasdk.api.shared_types import to_ISODate
from datetime import datetime


class EIAForecasts(Search):
    """EIA forecasts Endpoint, use this to search through Vortexa's EIA Forecasts data.

    The data includes:

    - `date`: date of the forecast
    - `forecast_fri`: Vortexa's data science based forecast of the EIA number to be published on the week
    - `value`: Actual EIA import/export numbers as published by the EIA Weekly Supply Estimates report
    - `stocks`: EIA stocks (kbl)
    - `cover`: Cover (days of Supply for the whole of the US, as published by the EIA Weekly Supply Estimates report)
    - `runs`: refinery runs (refiner “Percent Operable Utilization” as published by the EIA Weekly Supply Estimates report)

    """

    def __init__(self):
        Search.__init__(self, EIA_FORECASTS_RESOURCE)

    def search(
        self,
        preset: str = "padd1-gasoline-imports",
        filter_time_min: datetime = datetime(2020, 1, 1),
        filter_time_max: datetime = datetime(2020, 1, 31),
    ) -> EIAForecastResult:
        """
        Find EIA forecasts for a given preset and date range.

        # Arguments
            preset: Use to specify what geography and product information you would like to query.
            Preset can be: 'padd1-gasoline-imports', 'padd3-gasoline-imports', 'padd5-gasoline-imports',
            'us-gasoline-exports', 'padd1-crude-imports', 'padd3-crude-imports', 'padd5-crude-imports',
            'us-crude-exports', 'padd1-diesel-imports', 'padd3-diesel-imports', 'padd5-diesel-imports',
            'us-diesel-exports', 'padd1-jet-imports', 'padd5-jet-imports', 'us-jet-exports',
            'padd1-fueloil-imports', 'padd3-fueloil-imports', 'padd5-fueloil-imports' or 'us-fueloil-exports'

            filter_time_min: The UTC start date of the time filter

            filter_time_max: The UTC end date of the time filter

        # Returns
        List of EIA Forecast object matching selected 'preset'.

        # Examples

        Find PADD5 gasoline imports EIA forecasts from January 2019.
        ```python
        >>> from datetime import datetime
        >>> from vortexasdk import EIAForecasts
        >>> df = EIAForecasts().search(
        ...     preset="padd5-gasoline-imports",
        ...     filter_time_min=datetime(2020, 1, 1),
        ...     filter_time_max=datetime(2020, 1, 31)
        ... ).to_df()

        ```

        returns

        | date                     | forecast_fri     | value | stocks | cover | runs |
        | ------------------------ | ---------------- | ----- | ------ | ----- | ---- |
        | 2020-01-31T00:00:00.000Z | 454.96048964485  | 323   | 9541   | 26.5  | 65.9 |
        | 2020-01-24T00:00:00.000Z | 545.453497230504 | 579   | 10461  | 25.9  | 61.5 |
        | 2020-01-17T00:00:00.000Z | 510.289752707662 | 549   | 10325  | 25.2  | 64.7 |
        | 2020-01-10T00:00:00.000Z | 469.841470826967 |       |        |       |      |
        | 2020-01-03T00:00:00.000Z | 640.443229654771 |       |        |       |      |

        Some values can be NULL: value, stocks, cover, runs. It can happen when:

        - it's a very recent forecast, the Vortexa's data science based forecast (forecast_fri) is available but
          the complete EIA data isn't yet
        - it's an older forecast and the data is not available

        """
        search_params = {
            "preset": preset,
            "filter_time_min": to_ISODate(filter_time_min),
            "filter_time_max": to_ISODate(filter_time_max),
        }

        return EIAForecastResult(super().search(**search_params))



from dataclasses import dataclass
from typing import List, Optional, Tuple

from vortexasdk.api.serdes import FromDictMixin
from vortexasdk.api.shared_types import (
    EntityWithProbability,
    ID,
    IDLayer,
    IDNameLayer,
    Node,
)

Position = Tuple[float, float]


@dataclass(frozen=True)
class BoundingBox:
    """Polygon with list of bounding lon lat coords."""

    type: str
    coordinates: List[Position]


@dataclass(frozen=True)
class Geography(Node, FromDictMixin):
    """Represent a Geography reference record returned by the API."""

    id: ID
    name: str
    layer: List[str]
    exclusion_rule: List[IDNameLayer]
    hierarchy: List[IDLayer]
    bounding_box: Optional[BoundingBox]
    centre_point: Optional[Position]
    location: Optional[Position]


@dataclass(frozen=True)
class GeographyEntity(EntityWithProbability):
    """
    Represents a hierarchy tree of locational data.

    [Geography Entities Further Documentation](https://docs.vortexa.com/reference/intro-geography-entries)

    """



from unittest import TestCase

from scripts.validate_proposed_package_version import check_proposed_version_is_allowed


class TestVersionProposal(TestCase):
    def test_incremental_bump_is_allowed(self):
        check_proposed_version_is_allowed("0.18.0", "1.0.0")

    def test_incremental_minor_is_allowed(self):
        check_proposed_version_is_allowed("0.18.0", "0.19.0")

    def test_incremental_patch_is_allowed(self):
        check_proposed_version_is_allowed("0.18.0", "0.18.1")

    def test_prerelease_is_allowed(self):
        check_proposed_version_is_allowed("0.18.0", "0.19.1-foo")

    def test_non_semver_is_not_allowed(self):
        self.assertRaises(Exception, lambda: check_proposed_version_is_allowed("1.18.0", "foobar"))

    def test_old_major_is_not_allowed(self):
        self.assertRaises(Exception, lambda: check_proposed_version_is_allowed("1.18.0", "0.18.0"))

    def test_old_minor_is_not_allowed(self):
        self.assertRaises(Exception, lambda: check_proposed_version_is_allowed("0.18.0", "0.17.0"))

    def test_old_bump_is_not_allowed(self):
        self.assertRaises(Exception, lambda: check_proposed_version_is_allowed("0.18.1", "0.18.0"))



from dataclasses import dataclass
from typing import List

from vortexasdk.api.serdes import FromDictMixin
from vortexasdk.api.shared_types import IDName, EntityWithProbability


@dataclass(frozen=True)
class Corporation(IDName, FromDictMixin):
    """Represent a Corporation reference record returned by the API."""

    corporate_entity_type: List[str]
    ref_type: str
    leaf: bool
    parent: List[str]


@dataclass(frozen=True)
class CorporateEntity(EntityWithProbability):
    """
    Represents a relationship between a corporation and another entity like a vessel.

    [Corporate Entity Further Documentation](https://docs.vortexa.com/reference/intro-corporate-entities)


    """



from unittest import TestCase

import jsons

from vortexasdk.api.geography import GeographyEntity


class TestGeographyEntity(TestCase):
    def test_serialize(self):
        with open("tests/api/examples/geography_entity1.json", "r") as f:
            serialized = f.read()
            deserialized = jsons.loads(serialized, GeographyEntity)

            print(serialized)
            print(deserialized)

            expected = GeographyEntity(
                id="2aaad41b89dfad19e5668918018ae02695d7710bcbe5f2dc689234e8da492de3",
                layer="country",
                label="United Kingdom",
                source="model",
                probability=1,
            )

            assert expected == deserialized



from tests.testcases import TestCaseUsingRealAPI
from tests.timer import Timer
from vortexasdk import Products
from vortexasdk.endpoints.vessels import Vessels


class TestVesselsReal(TestCaseUsingRealAPI):
    def test_search_ids(self):
        ids = [
            "6d8a8f0863ca087204dd68e5fc3b6469a879829e6262856e34856aea3ca20509",
            "bf2b55bd31c709aa4cba91a3cc4111191c88c83753cbd285674c22150e42003e",
        ]

        vessels = Vessels().search(ids=ids).to_list()
        assert len(vessels) == 2

        print([x.name for x in vessels])

    def test_search_filters_vessel_class(self):
        vessel_classes = ["vlcc_plus", "aframax"]

        vessels = Vessels().search(vessel_classes=vessel_classes).to_list()

        actual = {x.vessel_class for x in vessels}

        assert actual == set(vessel_classes)

    def test_search_terms_are_combined_with_AND(self):
        aframax = set(
            v.id for v in Vessels().search(vessel_classes="aframax").to_list()
        )
        aframax_called_zhen = set(
            v.id
            for v in Vessels()
            .search(vessel_classes="aframax", term="zhen")
            .to_list()
        )

        assert aframax_called_zhen.issubset(aframax)

    def test_search_ids_dataframe(self):
        ids = [
            "6d8a8f0863ca087204dd68e5fc3b6469a879829e6262856e34856aea3ca20509",
            "bf2b55bd31c709aa4cba91a3cc4111191c88c83753cbd285674c22150e42003e",
        ]

        df = Vessels().search(ids=ids).to_df()
        assert list(df.columns) == ["id", "name", "imo", "vessel_class"]
        assert len(df) == 2

    def test_find_crude_vessels(self):
        crude = [
            p.id
            for p in Products().search("crude").to_list()
            if "group" in p.layer
        ]
        df = Vessels().search(vessel_product_types=crude).to_df()
        assert len(df) > 1000

    def test_load_all(self):
        all_products = Vessels().load_all()

        assert len(all_products) > 1000

    def test_search_load_all_vessels(self):
        with Timer("Search"):
            result = Vessels().search()

        with Timer("Serialize"):
            result.to_list()

        with Timer("Dataframe"):
            df = result.to_df()
            print(df.head())

        assert len(result) >= 1_000

    def test_search_is_case_agnostic(self):
        uppercase = {
            v.id for v in Vessels().search(vessel_classes="Suezmax").to_list()
        }
        lowercase = {
            v.id for v in Vessels().search(vessel_classes="suezmax").to_list()
        }

        assert uppercase == lowercase



from dataclasses import dataclass
from typing import List, Optional

from vortexasdk.api.corporation import CorporateEntity
from vortexasdk.api.id import ID
from vortexasdk.api.serdes import FromDictMixin
from vortexasdk.api.shared_types import (
    IDName,
    ISODate,
    Node,
    Tag,
    Scrubber,
    Flag,
)


@dataclass(frozen=True,)
class Vessel(Node, FromDictMixin):
    """
    Represent a Vessel reference record returned by the API.

    [Vessels Further Documentation](https://docs.vortexa.com/reference/GET/reference/vessels/%7Bid%7D)
    """

    related_names: List[str]
    mmsi: int

    tags: List[Tag]
    current_product_type: List

    vessel_class: str

    dead_weight: Optional[int] = None
    cubic_capacity: Optional[int] = None
    to_bow: Optional[str] = None
    to_stern: Optional[str] = None
    to_port: Optional[str] = None
    to_starboard: Optional[str] = None
    call_sign: Optional[str] = None
    year: Optional[int] = None
    imo: Optional[int] = None
    gross_tonnage: Optional[int] = None

    scrubber: Optional[Scrubber] = None
    flag: Optional[Flag] = None
    ice_class: Optional[str] = None
    propulsion: Optional[str] = None


@dataclass(frozen=True)
class VesselEntity(IDName):
    """
    A VesselEntity represents a vessel record used in CargoMovements and VesselMovements.

    [Vessel Entities Further Documentation](https://docs.vortexa.com/reference/intro-vessel-entities)
    """

    id: ID
    name: str
    mmsi: int
    imo: Optional[int]

    dwt: int

    vessel_class: str
    corporate_entities: List[CorporateEntity]
    tags: List[Tag]
    status: str
    year: Optional[int] = None

    start_timestamp: Optional[ISODate] = None

    cubic_capacity: Optional[int] = None
    voyage_id: Optional[str] = None
    fixture_fulfilled: Optional[bool] = None
    end_timestamp: Optional[ISODate] = None
    fixture_id: Optional[str] = None

    scrubber: Optional[Scrubber] = None
    flag: Optional[Flag] = None
    ice_class: Optional[str] = None
    propulsion: Optional[str] = None



import copy
import functools
import getpass
import os
from json import JSONDecodeError
from multiprocessing.pool import ThreadPool
from random import shuffle
from typing import Dict, List
import uuid

from requests import Response
from tqdm import tqdm

from vortexasdk.abstract_client import AbstractVortexaClient
from vortexasdk.api.id import ID
from vortexasdk.endpoints.endpoints import API_URL
from vortexasdk.logger import get_logger
from vortexasdk.retry_session import (
    retry_get,
    retry_post,
)
from vortexasdk.utils import filter_empty_values
from vortexasdk.version_utils import is_sdk_version_outdated
from vortexasdk.version import __version__
from vortexasdk import __name__ as sdk_pkg_name

logger = get_logger(__name__)


class VortexaClient(AbstractVortexaClient):
    """The API client responsible for calling Vortexa's Public API."""

    _DEFAULT_PAGE_LOAD_SIZE = int(1e4)
    _N_THREADS = 6
    _MAX_ALLOWED_TOTAL = int(1e6)

    def __init__(self, **kwargs):
        self.api_key = kwargs["api_key"]

    def get_reference(self, resource: str, id: ID) -> List[Dict]:
        """Lookup reference data."""
        url = self._create_url(f"{resource}/{id}")
        response = retry_get(url)
        return _handle_response(response)["data"]

    def search(self, resource: str, **data) -> List:
        """Search using `resource` using `**data` as filter params."""
        url = self._create_url(resource)
        payload = self._cleanse_payload(data)
        logger.info(f"Payload: {payload}")

        probe_response = _send_post_request(url, payload, size=1, offset=0)
        total = self._calculate_total(probe_response)

        if total > self._MAX_ALLOWED_TOTAL:
            raise Exception(
                f"Attempting to query too many records at once. Attempted records: {total}, Max allowed records: {self._MAX_ALLOWED_TOTAL} . "
                f"Try reducing the date range to return fewer records."
            )
        elif total == 1:
            # Only one page response, no need to send another request, so return flattened response
            return probe_response["data"]
        else:
            # Multiple pages available, create offsets and fetch all responses
            responses = self._process_multiple_pages(
                total=total, url=url, payload=payload, data=data
            )
            flattened = self._flatten_response(responses)
            assert len(flattened) == total, (
                f"Incorrect number of records returned from API. "
                f"Actual: {len(flattened)}, expected: {total}"
            )
            return flattened

    def _create_url(self, path: str) -> str:
        return (
            f"{API_URL}{path}?_sdk=python_v{__version__}&apikey={self.api_key}"
        )

    def _process_multiple_pages(
        self, total: int, url: str, payload: Dict, data: Dict
    ) -> List:
        size = data.get("size", 1000)
        offsets = list(range(0, total, size))
        shuffle(offsets)

        with tqdm(
            total=total, desc="Loading from API", disable=(len(offsets) == 1)
        ) as pbar:
            with ThreadPool(self._N_THREADS) as pool:
                logger.info(
                    f"{total} Results to retrieve."
                    f" Sending {len(offsets)}"
                    f" post requests in parallel using {self._N_THREADS} threads."
                )

                func = functools.partial(
                    _send_post_request_data,
                    url=url,
                    payload=payload,
                    size=size,
                    progress_bar=pbar,
                )

                return pool.map(func, offsets)

    @staticmethod
    def _cleanse_payload(payload: Dict) -> Dict:
        exclude_params = payload.get("exclude", {})
        payload["exclude"] = filter_empty_values(exclude_params)

        return filter_empty_values(payload)

    @staticmethod
    def _calculate_total(response: Dict) -> int:
        """ Get total number of pages, if total key does not exist, return 1 """
        return response.get("total", 1)

    @staticmethod
    def _flatten_response(response) -> List:
        return [x for y in response for x in y]


def _send_post_request_data(
    offset, url, payload, size, progress_bar: tqdm
) -> List:
    # noinspection PyBroadException
    try:
        progress_bar.update(size)
    except Exception:
        logger.warn("Could not update progress bar")

    dict_response = _send_post_request(url, payload, size, offset)

    return dict_response.get("data", [])


def _send_post_request(url, payload, size, offset) -> Dict:
    logger.debug(f"Sending post request, offset: {offset}, size: {size}")

    payload_with_offset = copy.deepcopy(payload)

    payload_with_offset["offset"] = offset
    payload_with_offset["cm_offset"] = offset
    payload_with_offset["size"] = size
    payload_with_offset["cm_size"] = size

    response = retry_post(url, json=payload_with_offset)

    return _handle_response(response, payload_with_offset)


def _handle_response(response: Response, payload: Dict = None) -> Dict:
    if not response.ok:
        logger.error(response.reason)
        logger.error(response.status_code)
        logger.error(response)

        # noinspection PyBroadException
        try:
            logger.error(response.json())
            message = response.json()["message"]
        except Exception:
            message = ""
            pass

        logger.error(f"payload: {payload}")
        error = f"[{response.status_code} {response.reason}]"

        raise ValueError(f"{error} {message}")

    else:
        try:
            json = response.json()
        except JSONDecodeError:
            logger.error("Could not decode response")
            json = {}
        except Exception as e:
            logger.error(e)
            json = {}

    return json


__client__ = None


def default_client() -> VortexaClient:
    """Instantiate VortexaClient as global variable."""
    global __client__

    if __client__ is None:
        __client__ = create_client()

    return __client__


def create_client() -> VortexaClient:
    """Create new VortexaClient."""
    logger.info("Creating new VortexaClient")

    api_key = _load_api_key()
    verify_api_key_format(api_key)
    _warn_user_if_sdk_version_outdated()

    return VortexaClient(api_key=api_key)


def set_client(client) -> None:
    """Set the global client, used by all endpoints."""
    global __client__
    __client__ = client
    logger.debug(
        f"global __client__ has been set {__client__.__class__.__name__} \n"
    )


def _warn_user_if_sdk_version_outdated() -> None:
    """Warn users if their SDK version is outdated"""
    try:
        latest_sdk_version, sdk_outdated_check = is_sdk_version_outdated()
        if sdk_outdated_check:
            logger.warning(
                f"You are using {sdk_pkg_name} version {__version__}, however version {latest_sdk_version} is available.\n"
                f"You should consider upgrading via the 'pip install {sdk_pkg_name} --upgrade' command."
            )
    except Exception as e:
        logger.warning(
            f"Outdated SDK version check could not be completed. \n"
            f"Got an exception: {e}"
        )


def _load_api_key():
    """Read API Key from environment variables else user input"""
    try:
        return os.environ["VORTEXA_API_KEY"]
    except KeyError:
        return getpass.getpass("Vortexa API Key: ")
    except Exception:
        raise KeyError(
            "You must either set the VORTEXA_API_KEY environment variable, or interactively enter your Vortexa API key."
            " Your API key can be found at https://docs.vortexa.com"
        )


def verify_api_key_format(api_key: str) -> None:
    """Verify that the api_key is a valid UUID string"""
    try:
        uuid.UUID(api_key)
    except ValueError:
        raise ValueError(
            "Incorrect API key set. The Vortexa API key must be of the form xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx "
            " Your API key can be found at https://docs.vortexa.com"
        )



"""Geographies Endpoint."""
from typing import Dict, List, Union

from vortexasdk.api import ID
from vortexasdk.endpoints.endpoints import GEOGRAPHIES_REFERENCE
from vortexasdk.endpoints.geographies_result import GeographyResult
from vortexasdk.operations import Reference, Search
from vortexasdk.utils import convert_values_to_list


class Geographies(Reference, Search):
    """Geographies endpoint."""

    def __init__(self):
        Reference.__init__(self, GEOGRAPHIES_REFERENCE)
        Search.__init__(self, GEOGRAPHIES_REFERENCE)

    def load_all(self) -> GeographyResult:
        """Load all geographies."""
        return self.search()

    def search(
        self,
        term: Union[str, List[str]] = None,
        exact_term_match: bool = False,
    ) -> GeographyResult:
        """
        Find all geographies matching given search terms.

        # Arguments
            term: The geography name (or names) we're filtering on

            exact_term_match: Search on only exact term matches, or allow similar matches.
                e.g. When searching for "China" with `exact_term_match=False`, then the SDK will yield geographies named
                ['China', 'South China', 'China Energy Services Ningbo'...] etc. When `exact_term_match=True`,
                the SDK will only yield the geography named `China`.


        # Returns
        List of geographies matching `term`


        # Examples

        Find all geographies with `portsmouth` in the name.
        ```python
        >>> from vortexasdk import Geographies
        >>> [x.name for x in Geographies().search(term="portsmouth").to_list()]
        ['Portsmouth [GB]', 'Portsmouth, NH [US]']

        ```

        Search multiple geography terms
        ```python
        >>> df = Geographies().search(term=["Liverpool", "Southampton"]).to_df()

        ```
        returns

        |    | id                | name                   | layer        |
        |---:|:------------------|:-----------------------|:-------------|
        |  0 | b63d8f625669fd... | Liverpool [GB]         | ['port']     |
        |  1 | 0cb7d4566de0f2... | Southampton [GB]       | ['port']     |
        |  2 | 8b4273e3181f2d... | Liverpool Docks        | ['terminal'] |
        |  3 | 98c50b0d2ee2b1... | Liverpool Bulk Liquids | ['terminal'] |
        """
        api_params = convert_values_to_list({"term": term})

        return GeographyResult(
            super().search(exact_term_match=exact_term_match, **api_params)
        )

    def reference(self, id: ID) -> Dict:
        """
        Perform a geography lookup.

        # Arguments
            id: Geography ID to lookup

        # Returns
         Geography matching the ID

        # Further Documentation:
        [VortexaAPI Geography Reference](https://docs.vortexa.com/reference/GET/reference/geographies/%7Bid%7D)

        """
        return super().reference(id)



from unittest import TestCase

from vortexasdk.client import (
    VortexaClient,
    verify_api_key_format,
    _handle_response,
)

from requests.models import Response


class TestClient(TestCase):
    def test__cleanse_payload(self):
        payload = {"a": [], "b": None, "d": 3, "c": [1, 2]}

        cleansed = VortexaClient._cleanse_payload(payload)

        assert cleansed == {"c": [1, 2], "d": 3}

    def test__cleanse_payload_removes_exclude_empties(self):
        payload = {
            "layer1_k1": [],
            "layer1_k2": None,
            "layer1_k3": 3,
            "exclude": {"layer2_k1": [], "layer2_k2": [1]},
        }

        cleansed = VortexaClient._cleanse_payload(payload)

        assert cleansed == {"layer1_k3": 3, "exclude": {"layer2_k2": [1]}}

    def test_raised_error_for_invalid_uuid(self):
        self.assertRaises(
            ValueError, lambda: verify_api_key_format("invalid_key")
        )

    def test_accepts_valid_uuid(self):
        sample_valid_key = "123e4567-e89b-12d3-a456-426614174000"
        verify_api_key_format(sample_valid_key)

    def test_raised_error_for_response_error(self):
        badResponse = Response()
        badResponse.code = "bad request"
        badResponse.reason = "bad request"
        badResponse.status_code = 400
        self.assertRaises(ValueError, lambda: _handle_response(badResponse))



from datetime import datetime

from tests.testcases import TestCaseUsingMockAPI
from vortexasdk.endpoints.cargo_timeseries import CargoTimeSeries


class TestCargoTimeSeries(TestCaseUsingMockAPI):
    def test_search_returns_all_days(self):
        start = datetime(2018, 10, 1)
        end = datetime(2019, 11, 1)

        df = (
            CargoTimeSeries()
            .search(
                filter_activity="loading_state",
                filter_time_min=start,
                filter_time_max=end,
            )
            .to_df()
        )

        assert len(df) == 3



from typing import List, Union

from vortexasdk.endpoints.products import Products
from vortexasdk.api import ID
from vortexasdk.conversions.conversions import _convert_to_ids


def convert_to_product_ids(
    ids_or_names_list: List[Union[ID, str]]
) -> List[ID]:
    """
    Convert a mixed list of names or IDs to vessel ids.

    # Example
    ```
    >>> convert_to_product_ids(["crude"])
    [...]

    ```
    """
    return _convert_to_ids(ids_or_names_list, Products())



from datetime import datetime
from unittest import TestCase

from vortexasdk.api.shared_types import to_ISODate


class TestToISODate(TestCase):
    def test_to_ISODate(self):
        assert to_ISODate(datetime(2018, 12, 31)) == "2018-12-31T00:00:00.000Z"

    def test_to_ISODate_with_time(self):
        assert (
            to_ISODate(datetime(2018, 12, 31, 23, 59, 59, 123456))
            == "2018-12-31T23:59:59.123Z"
        )



from typing import List, Dict

import jsons

from vortexasdk.api import ID
from vortexasdk.abstract_client import AbstractVortexaClient
from vortexasdk.endpoints.endpoints import (
    CORPORATIONS_REFERENCE,
    VESSELS_REFERENCE,
    PRODUCTS_REFERENCE,
    ATTRIBUTES_REFERENCE,
    VESSEL_MOVEMENTS_RESOURCE,
    CARGO_TIMESERIES_RESOURCE,
)


def _read(example_file) -> List[Dict]:
    with open(f"tests/api/examples/{example_file}", "r") as f:
        return jsons.loads(f.read(), List)


example_corporations: List[Dict] = _read("corporations.json")
example_vessels: List[Dict] = _read("vessels.json")
example_products: List[Dict] = _read("products.json")
example_attributes: List[Dict] = _read("attributes.json")
example_vessel_movements: List[Dict] = _read("vessel_movements.json")
example_cargo_movements: List[Dict] = _read("vessel_movements.json")
example_time_series = _read("cargo_time_series.json")


class MockVortexaClient(AbstractVortexaClient):
    _results = {
        CORPORATIONS_REFERENCE: example_corporations,
        ATTRIBUTES_REFERENCE: example_attributes,
        VESSELS_REFERENCE: example_vessels,
        PRODUCTS_REFERENCE: example_products,
        VESSEL_MOVEMENTS_RESOURCE: example_vessel_movements,
        CARGO_TIMESERIES_RESOURCE: example_time_series,
    }

    def get_reference(self, resource: str, id: ID) -> List[Dict]:
        entities = MockVortexaClient._results[resource]
        return [e for e in entities if e["id"] == id]

    def search(self, resource: str, **data) -> List:
        return MockVortexaClient._results[resource]



"""Cargo Movements Endpoint."""
from datetime import datetime
from typing import List, Union

from vortexasdk.api import ID
from vortexasdk.api.shared_types import to_ISODate
from vortexasdk.endpoints.cargo_movements_result import CargoMovementsResult
from vortexasdk.endpoints.endpoints import CARGO_MOVEMENTS_RESOURCE
from vortexasdk.logger import get_logger
from vortexasdk.operations import Search
from vortexasdk.utils import convert_to_list

logger = get_logger(__name__)


class CargoMovements(Search):
    """
    Cargo Movements Endpoint, use this to search through Vortexa's cargo movements.

    A detailed explanation of Cargo/Vessel Movements can be found [here](https://docs.vortexa.com/reference/intro-movement-difference).
    """

    _MAX_PAGE_RESULT_SIZE = 500

    def __init__(self):
        Search.__init__(self, CARGO_MOVEMENTS_RESOURCE)

    def search(
        self,
        filter_activity: str,
        filter_time_min: datetime = datetime(2019, 10, 1, 0),
        filter_time_max: datetime = datetime(2019, 10, 1, 1),
        cm_unit: str = "b",
        filter_charterers: Union[ID, List[ID]] = None,
        filter_destinations: Union[ID, List[ID]] = None,
        filter_origins: Union[ID, List[ID]] = None,
        filter_owners: Union[ID, List[ID]] = None,
        filter_products: Union[ID, List[ID]] = None,
        filter_vessels: Union[ID, List[ID]] = None,
        filter_storage_locations: Union[ID, List[ID]] = None,
        filter_ship_to_ship_locations: Union[ID, List[ID]] = None,
        filter_waypoints: Union[ID, List[ID]] = None,
        filter_vessel_age_min: int = None,
        filter_vessel_age_max: int = None,
        filter_vessel_scrubbers: str = "disabled",
        filter_vessel_flags: Union[ID, List[ID]] = None,
        filter_vessel_ice_class: Union[ID, List[ID]] = None,
        filter_vessel_propulsion: Union[ID, List[ID]] = None,
        exclude_origins: Union[ID, List[ID]] = None,
        exclude_destinations: Union[ID, List[ID]] = None,
        exclude_products: Union[ID, List[ID]] = None,
        exclude_vessels: Union[ID, List[ID]] = None,
        exclude_charterers: Union[ID, List[ID]] = None,
        exclude_owners: Union[ID, List[ID]] = None,
        exclude_vessel_flags: Union[ID, List[ID]] = None,
        exclude_vessel_ice_class: Union[ID, List[ID]] = None,
        exclude_vessel_propulsion: Union[ID, List[ID]] = None,
        disable_geographic_exclusion_rules: bool = None,
        timeseries_activity_time_span_min: int = None,
        timeseries_activity_time_span_max: int = None,
    ) -> CargoMovementsResult:
        """

        Find CargoMovements matching the given search parameters.

        # Arguments
            filter_activity: Movement activity on which to base the time filter. Must be one of ['loading_state',
             'loading_start', 'loading_end', 'identified_for_loading_state', 'unloading_state', 'unloading_start',
              'unloading_end', 'unloaded_state', 'storing_state', 'storing_start', 'storing_end', 'transiting_state',
               'any_activity'].

            filter_time_min: The UTC start date of the time filter.

            filter_time_max: The UTC end date of the time filter.

            cm_unit: Unit of measurement. Enter 'b' for barrels or 't' for tonnes.

            filter_charterers: A charterer ID, or list of charterer IDs to filter on.

            filter_destinations: A geography ID, or list of geography IDs to filter on.

            filter_origins: A geography ID, or list of geography IDs to filter on.

            filter_owners: An owner ID, or list of owner IDs to filter on.

            filter_products: A product ID, or list of product IDs to filter on.

            filter_vessels: A vessel ID, or list of vessel IDs to filter on.

            filter_storage_locations: A geography ID, or list of geography IDs to filter on.

            filter_ship_to_ship_locations: A geography ID, or list of geography IDs to filter on.

            filter_waypoints: A geography ID, or list of geography IDs to filter on.

            filter_vessel_age_min: A number between 1 and 100 (representing years).

            filter_vessel_age_max: A number between 1 and 100 (representing years).

            filter_vessel_scrubbers: Either inactive 'disabled', or included 'inc' or excluded 'exc'.

            filter_vessel_flags: A vessel flag, or list of vessel flags to filter on.

            filter_vessel_ice_class: An attribute ID, or list of attribute IDs to filter on.

            filter_vessel_propulsion: An attribute ID, or list of attribute IDs to filter on.

            exclude_origins: A geography ID, or list of geography IDs to exclude.

            exclude_destinations: A geography ID, or list of geography IDs to exclude.

            exclude_products: A product ID, or list of product IDs to exclude.

            exclude_vessels: A vessel ID, or list of vessel IDs to exclude.

            exclude_charterers: A charterer ID, or list of charterer IDs to exclude.

            exclude_owners: An owner ID, or list of owner IDs to exclude.

            exclude_vessel_flags: A geography ID, or list of geography IDs to exclude.

            exclude_vessel_ice_class: An attribute ID, or list of attribute IDs to exclude.

            exclude_vessel_propulsion: An attribute ID, or list of attribute IDs to exclude.

            disable_geographic_exclusion_rules: This controls a popular industry term "intra-movements" and determines
             the filter behaviour for cargo leaving then entering the same geographic area.

            timeseries_activity_time_span_min: The minimum amount of time in milliseconds accounted for in a time series
             activity. Can be used to request long-term floating storage. For example, to only return floating storage
             movements that occured for _more_ than 14 days enter
             `timeseries_activity_time_span_min=1000 * 60 * 60 * 24 * 14` in conjunction with
             `filter_activity='storing_state'`.

            timeseries_activity_time_span_max: The maximum amount of time in milliseconds accounted for in a time series
             activity. Can be used to request short-term floating storage. For example, to only return floating storage
             movements that occured for _less_ than 14 days enter
             `timeseries_activity_time_span_max=1000 * 60 * 60 * 24 * 14`
             in conjunction with `filter_activity='storing_state'`.

        # Returns
        `CargoMovementsResult`, containing all the cargo movements matching the given search terms.


        # Example

        * _Which cargoes were loaded from Rotterdam on the morning of 1st December 2018?_


        ```python
        >>> from vortexasdk import CargoMovements, Geographies
        >>> rotterdam = [g.id for g in Geographies().search("rotterdam").to_list() if "port" in g.layer]
        >>> search_result = CargoMovements().search(
        ...    filter_origins=rotterdam,
        ...    filter_activity='loading_state',
        ...    filter_time_min=datetime(2018, 12, 1),
        ...    filter_time_max=datetime(2018, 12, 1, 12))
        >>> df = search_result.to_df(columns=['product.grade.label', 'product.group.label', 'vessels.0.vessel_class'])

        ```
        |    | product.group.label   | product.grade.label             | vessels.0.vessel_class   |
        |---:|:----------------------|:--------------------------------|:-------------------------|
        |  0 | Clean products        | Pygas                           | general_purpose          |
        |  1 | Clean products        | Chemicals                       | tiny_tanker              |
        |  2 | Clean products        | Chemicals                       | tiny_tanker              |
        |  3 | Dirty products        | Low Sulphur VGO (LSVGO)         | general_purpose          |
        |  4 | Clean products        | ULSD (Ultra Low Sulphur Diesel) | general_purpose          |
        |  5 | Clean products        | Chemicals                       | tiny_tanker              |
        |  6 | Clean products        | Finished Gasoline               | handymax                 |

        * _Which VLCC cargoes passed through the Suez canal en route to China?_

        Note here we include vessels.0..., vessels.1..., vessels.2... columns.
        This lets us view all vessels present in any STS operations.

        ```python
        >>> from vortexasdk import CargoMovements, Geographies, Vessels
        >>> suez = [g.id for g in Geographies().search("suez").to_list()]
        >>> china = [g.id for g in Geographies().search("china").to_list() if "country" in g.layer]
        >>> vlccs = [v.id for v in Vessels().search(vessel_classes="vlcc_plus").to_list()]
        >>> cargo_movement_search_result = CargoMovements().search(
        ...    filter_destinations=china,
        ...    filter_activity="loading_state",
        ...    filter_waypoints=suez,
        ...    filter_vessels=vlccs,
        ...    filter_time_min=datetime(2018, 12, 1),
        ...    filter_time_max=datetime(2018, 12, 1))
        >>> cols = ['vessels.0.name', 'vessels.0.vessel_class', 'vessels.1.name', 'vessels.1.vessel_class',  'vessels.2.name', 'vessels.2.vessel_class', 'product.group.label', 'quantity']
        >>> cargo_movements_df = cargo_movement_search_result.to_df(columns=cols)

        ```

        |    | vessels.0.name   | vessels.0.vessel_class   | vessels.1.name   | vessels.1.vessel_class   | vessels.2.name   | vessels.2.vessel_class   | product.group.label   |   quantity |
        |---:|:-----------------|:-------------------------|:-----------------|:-------------------------|:-----------------|:-------------------------|:----------------------|-----------:|
        |  0 | MINERVA MARINA   | suezmax                  | COSGLORY LAKE    | vlcc_plus                | nan              | nan                      | Crude                 |     700614 |
        |  1 | BUKHA            | vlcc_plus                | nan              | nan                      | nan              | nan                      | Crude                 |    1896374 |
        |  2 | ATHENIAN FREEDOM | vlcc_plus                | nan              | nan                      | nan              | nan                      | Crude                 |     183537 |
        |  3 | ATINA            | suezmax                  | DONAT            | suezmax                  | DS VISION        | vlcc_plus                | Crude                 |     896773 |
        |  4 | MINERVA MARINA   | suezmax                  | COSGLORY LAKE    | vlcc_plus                | nan              | nan                      | Crude                 |     405724 |
        |  5 | MASAL            | suezmax                  | EKTA             | vlcc_plus                | nan              | nan                      | Crude                 |     997896 |
        |  6 | ATHENIAN FREEDOM | vlcc_plus                | nan              | nan                      | nan              | nan                      | Crude                 |     120812 |

        [Cargo Movements Endpoint Further Documentation](https://docs.vortexa.com/reference/POST/cargo-movements/search)

        """
        exclude_params = {
            "filter_origins": convert_to_list(exclude_origins),
            "filter_destinations": convert_to_list(exclude_destinations),
            "filter_products": convert_to_list(exclude_products),
            "filter_vessels": convert_to_list(exclude_vessels),
            "filter_charterers": convert_to_list(exclude_charterers),
            "filter_owners": convert_to_list(exclude_owners),
            "filter_vessel_flags": convert_to_list(exclude_vessel_flags),
            "filter_vessel_ice_class": convert_to_list(
                exclude_vessel_ice_class
            ),
            "filter_vessel_propulsion": convert_to_list(
                exclude_vessel_propulsion
            ),
        }

        api_params = {
            "filter_activity": filter_activity,
            "filter_time_min": to_ISODate(filter_time_min),
            "filter_time_max": to_ISODate(filter_time_max),
            "timeseries_activity_time_span_min": timeseries_activity_time_span_min,
            "timeseries_activity_time_span_max": timeseries_activity_time_span_max,
            "cm_unit": cm_unit,
            "filter_charterers": convert_to_list(filter_charterers),
            "filter_owners": convert_to_list(filter_owners),
            "filter_products": convert_to_list(filter_products),
            "filter_vessels": convert_to_list(filter_vessels),
            "filter_destinations": convert_to_list(filter_destinations),
            "filter_origins": convert_to_list(filter_origins),
            "filter_storage_locations": convert_to_list(
                filter_storage_locations
            ),
            "filter_ship_to_ship_locations": convert_to_list(
                filter_ship_to_ship_locations
            ),
            "filter_waypoints": convert_to_list(filter_waypoints),
            "filter_vessel_age_min": filter_vessel_age_min,
            "filter_vessel_age_max": filter_vessel_age_max,
            "filter_vessel_scrubbers": filter_vessel_scrubbers,
            "filter_vessel_flags": convert_to_list(filter_vessel_flags),
            "filter_vessel_ice_class": convert_to_list(
                filter_vessel_ice_class
            ),
            "filter_vessel_propulsion": convert_to_list(
                filter_vessel_propulsion
            ),
            "exclude": exclude_params,
            "disable_geographic_exclusion_rules": disable_geographic_exclusion_rules,
            "size": self._MAX_PAGE_RESULT_SIZE,
        }

        return CargoMovementsResult(super().search(**api_params))
